/*
 * arch/ppc/kernel/mpc5xxx_pci.c
 *
 * PCI Support for MPC5xxx Processors
 *
 * 2003 (c) Wolfgang Denk, DENX Software Engineering, <wd@denx.de>.  This
 * file is licensed under the terms of the GNU General Public License
 * version 2.  This program is licensed "as is" without any warranty of
 * any kind, whether express or implied.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/mpc5xxx.h>

#define MPC5xxx_PCI_CAR		0xf0000df8
#define CONFIG_PCI_IO_PHYS	0x50000000

static int mpc5xxx_read_config_dword(struct pci_dev *dev,
					int offset, u32* value)
{
	struct pci_controller *hose = dev->sysdata;

	*hose->cfg_addr = (1 << 31) | (dev->devfn << 8) |
			(dev->bus->number << 16) | (offset & 0xfc);
	eieio();
	udelay(10);
	*value = in_le32((u32 *)hose->cfg_data);
	eieio();
	*hose->cfg_addr = 0;
	udelay(10);
	return 0;
}

static int mpc5xxx_write_config_dword(struct pci_dev *dev,
					int offset, u32 value)
{
	struct pci_controller *hose = dev->sysdata;

	*hose->cfg_addr = (1 << 31) | (dev->devfn << 8) |
			(dev->bus->number << 16) | (offset & 0xfc);
	eieio();
	udelay(10);
	out_le32((u32 *)hose->cfg_data, value);
	eieio();
	*hose->cfg_addr = 0;
	udelay(10);
	return 0;
}

#define PCI_READ_VIA_DWORD_OP(size, type, off_mask)			\
int mpc5xxx_read_config_##size##_via_dword(struct pci_dev *dev,		\
					int offset, type val)		\
{									\
	u32 val32;							\
									\
	if (mpc5xxx_read_config_dword(dev, offset & 0xfc, &val32) < 0)	\
		return -1;						\
									\
	*val = (val32 >> ((offset & (int)off_mask) * 8));		\
									\
	return 0;							\
}

#define PCI_WRITE_VIA_DWORD_OP(size, type, off_mask, val_mask)		\
int mpc5xxx_write_config_##size##_via_dword(struct pci_dev *dev,	\
						int offset, type val)	\
{									\
	u32 val32, mask, ldata;						\
									\
	if (mpc5xxx_read_config_dword(dev, offset & 0xfc, &val32) < 0)	\
		return -1;						\
									\
	mask = val_mask;						\
	ldata = (((unsigned long)val) & mask) <<			\
					((offset & (int)off_mask) * 8);	\
	mask <<= ((mask & (int)off_mask) * 8);				\
	val32 = (val32 & ~mask) | ldata;				\
									\
	if (mpc5xxx_write_config_dword(dev, offset & 0xfc, val32) < 0)	\
		return -1;						\
									\
	return 0;							\
}

PCI_READ_VIA_DWORD_OP(byte, u8 *, 0x03)
PCI_READ_VIA_DWORD_OP(word, u16 *, 0x02)
PCI_WRITE_VIA_DWORD_OP(byte, u8, 0x03, 0x000000ff)
PCI_WRITE_VIA_DWORD_OP(word, u16, 0x02, 0x0000ffff)

static struct pci_ops mpc5xxx_pci_ops =
{
	mpc5xxx_read_config_byte_via_dword,
	mpc5xxx_read_config_word_via_dword,
	mpc5xxx_read_config_dword,
	mpc5xxx_write_config_byte_via_dword,
	mpc5xxx_write_config_word_via_dword,
	mpc5xxx_write_config_dword
};

static void __init
mpc5xxx_init_pci(void)
{
	struct mpc5xxx_pci *pci = (struct mpc5xxx_pci *) MPC5xxx_PCI;
	struct mpc5xxx_gpio *gpio = (struct mpc5xxx_gpio *) MPC5xxx_GPIO;
	struct mpc5xxx_xlb *xlb = (struct mpc5xxx_xlb *) MPC5xxx_XLB;

	/* enable PCI support */
	out_be32(&gpio->port_config,
			in_be32(&gpio->port_config) & ~MPC5xxx_GPIO_PCI_DIS);

	/* enable PCI master and memory decode */
	out_be32(&pci->scr, in_be32(&pci->scr) | 
			PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY);

	/* set max latency */
	out_be32(&pci->cr1, in_be32(&pci->cr1) | 0xf800);

	/* set cache line size */
	out_be32(&pci->cr1, (in_be32(&pci->cr1) & ~0xf) | (32/4));

	/* map MBAR to PCI space */
	out_be32(&pci->bar0, MPC5xxx_MBAR);
	out_be32(&pci->tbatr0, MPC5xxx_MBAR | 1);

	/* map ram to PCI space */
	out_be32(&pci->bar1, MPC5xxx_PCI_DRAM_BASE | (1 << 3));
	out_be32(&pci->tbatr1, MPC5xxx_PCI_PHY_MEM_BASE | 1);

	/* enable ram snooping for 1GB window */
	out_be32(&xlb->config, in_be32(&xlb->config) | MPC5xxx_XLB_CFG_SNOOP);
	out_be32(&xlb->snoop_window, MPC5xxx_PCI_PHY_MEM_BASE | 0x1d);

	/* park xlb on pci */
	out_be32(&xlb->config, in_be32(&xlb->config) & ~((7<<8) | (3<<5)));
	out_be32(&xlb->config, in_be32(&xlb->config) |   (3<<8) | (3<<5) );

	/* disable pci controller interrupts */
	out_be32(&pci->gscr, in_be32(&pci->gscr) & ~(7 << 12));
	out_be32(&pci->icr,  in_be32(&pci->icr)  & ~(7 << 24));

	out_be32(&pci->iw0btar, PCI_WINDOW_TRANSLATION(
					MPC5xxx_PROC_PCI_MEM_START,
					MPC5xxx_PROC_PCI_MEM_END,
					MPC5xxx_PCI_MEM_START,
					MPC5xxx_PCI_MEM_END));
	out_be32(&pci->iw1btar, PCI_WINDOW_TRANSLATION(
					MPC5xxx_PROC_PCI_IO_START,
					MPC5xxx_PROC_PCI_IO_END,
					MPC5xxx_PCI_IO_START,
					MPC5xxx_PCI_IO_END));
	out_be32(&pci->iw2btar, 0);

	out_be32(&pci->iwcr, PCI_WINDOW_CONTROL(
				(
					MPC5xxx_INIT_WINDOW_ENABLE |
					MPC5xxx_INIT_WINDOW_MEM |
					MPC5xxx_INIT_WINDOW_READ
				),
				(
					MPC5xxx_INIT_WINDOW_ENABLE |
					MPC5xxx_INIT_WINDOW_IO
				),
				MPC5xxx_INIT_WINDOW_DISABLE));

	/* reset pci bus */
	out_be32(&pci->gscr, in_be32(&pci->gscr) | 1);
	udelay(1000);
	out_be32(&pci->gscr, in_be32(&pci->gscr) & ~1);
}

void __init
mpc5xxx_find_bridges(void)
{
	extern int pci_assign_all_busses;
	struct pci_controller *hose;

	mpc5xxx_init_pci();

	hose = pcibios_alloc_controller();

	pci_assign_all_busses = 1;

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno = 0xff;

	hose->io_space.start = MPC5xxx_PCI_IO_START;
	hose->io_space.end = MPC5xxx_PCI_IO_END;
	hose->io_base_virt = (void *)isa_io_base;

	pci_init_resource(&hose->io_resource,
		MPC5xxx_PCI_IO_START,
		MPC5xxx_PCI_IO_END,
		IORESOURCE_IO,
		"PCI host bridge");

	hose->mem_space.start = MPC5xxx_PCI_MEM_START;
	hose->mem_space.end = MPC5xxx_PCI_MEM_END;

	pci_init_resource(&hose->mem_resources[0],
		MPC5xxx_PCI_MEM_START,
		MPC5xxx_PCI_MEM_END,
		IORESOURCE_MEM,
		"PCI host bridge");

	hose->ops = &mpc5xxx_pci_ops;

	hose->cfg_addr = (unsigned int *) ioremap(MPC5xxx_PCI_CAR, 4);
	hose->cfg_data = (unsigned int *) ioremap(CONFIG_PCI_IO_PHYS, 4);

	hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);
}
