/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/dprobes.h>
#include <linux/hook.h>

void dprobes_hooks_asm_code_start(void) { }

USE_HOOK(DO_TRAP);
USE_HOOK(DO_IABR);
USE_HOOK(DO_PGCK);
USE_HOOK(DO_DEBUG);
USE_HOOK(DO_PAGE_FAULT);
USE_HOOK(DPROBES_CALLK);

static int do_trap_exit(hook_rec_t * hookrec, struct pt_regs * regs)
{
	dp_handle_fault(regs);
	return HOOK_CONTINUE;
}	


static int do_iabr_exit(hook_rec_t * hookrec, int *rc, struct pt_regs * regs)
{
	if ((*rc = dp_trap(regs, DP_PROBE_BREAKPOINT, 0, regs->nip)))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_pgck_exit(hook_rec_t * hookrec, int *rc, struct pt_regs * regs)
{
	if ((*rc = dp_trap(regs, DP_PROBE_BREAKPOINT, 0, regs->nip)))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_debug_exit(hook_rec_t * hookrec, struct pt_regs * regs)
{
	if (dp_trap1(regs))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

static int do_page_fault_exit(hook_rec_t * hookrec, struct pt_regs * regs)
{
	if (dp_pf(regs))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}

int initialise_asm_dprobes_hooks(void) 
{
	int retval;

	if ((retval = hook_initialise(DO_TRAP)))
		goto err;
	if ((retval = hook_initialise(DO_PGCK)))
		goto err1;
	if ((retval = hook_initialise(DO_IABR)))
		goto err2;
	if ((retval = hook_initialise(DO_DEBUG)))
		goto err3;
	if ((retval = hook_initialise(DO_PAGE_FAULT)))
		goto err4;
	if ((retval = hook_initialise(DPROBES_CALLK)))
		goto err5;
	return 0;
	
err5:	hook_terminate(DO_PAGE_FAULT, 0);
err4:	hook_terminate(DO_DEBUG, 0);
err3:	hook_terminate(DO_IABR, 0);
err2:	hook_terminate(DO_PGCK, 0);
err1:	hook_terminate(DO_TRAP, 0);
err:	return retval;
}

int terminate_asm_dprobes_hooks(void)
{
	hook_terminate(DPROBES_CALLK, 0);
	hook_terminate(DO_PAGE_FAULT, 0);
	hook_terminate(DO_DEBUG, 0);
	hook_terminate(DO_IABR, 0);
	hook_terminate(DO_PGCK, 0);
	hook_terminate(DO_TRAP, 0);
	return 0;
}
hook_rec_t do_trap_hook_rec;
hook_rec_t do_iabr_hook_rec;
hook_rec_t do_pgck_hook_rec;
hook_rec_t do_debug_hook_rec;
hook_rec_t do_page_fault_hook_rec;

int register_asm_dprobes_hooks(void)
{
	int retval;
	
	do_trap_hook_rec.hook_exit = do_trap_exit;
	do_trap_hook_rec.hook_exit_name = "dp_do_trap_exit";
	if ((retval = hook_exit_register(DO_TRAP, &do_trap_hook_rec, 0)))
		goto err;

	do_pgck_hook_rec.hook_exit = do_pgck_exit;
	do_pgck_hook_rec.hook_exit_name = "dp_do_pgck_exit";
	if ((retval = hook_exit_register(DO_PGCK, &do_pgck_hook_rec, 0)))
		goto err1;

	do_iabr_hook_rec.hook_exit = do_iabr_exit;
	do_iabr_hook_rec.hook_exit_name = "dp_do_iabr_exit";
	if ((retval = hook_exit_register(DO_IABR, &do_iabr_hook_rec, 0)))
		goto err2;

	do_debug_hook_rec.hook_exit = do_debug_exit;
	do_debug_hook_rec.hook_exit_name = "dp_do_debug_exit";
	if ((retval = hook_exit_register(DO_DEBUG, &do_debug_hook_rec, 0)))
		goto err3;

	do_page_fault_hook_rec.hook_exit = do_page_fault_exit;
	do_page_fault_hook_rec.hook_exit_name = "dp_do_page_fault_exit";
	if ((retval = hook_exit_register(DO_PAGE_FAULT, &do_page_fault_hook_rec, 0)))
		goto err4;
	return 0;
	
err4:	hook_exit_deregister(&do_debug_hook_rec);
err3:	hook_exit_deregister(&do_iabr_hook_rec);
err2:	hook_exit_deregister(&do_pgck_hook_rec);
err1:	hook_exit_deregister(&do_trap_hook_rec);
err:	return retval;

}

int deregister_asm_dprobes_hooks(void)
{
	hook_exit_deregister(&do_page_fault_hook_rec);
	hook_exit_deregister(&do_debug_hook_rec);
	hook_exit_deregister(&do_iabr_hook_rec);
	hook_exit_deregister(&do_pgck_hook_rec);
	hook_exit_deregister(&do_trap_hook_rec);
	return 0;
}

int arm_all_asm_dprobes_hooks(void)
{
	int retval;
	
	if ((retval = hook_exit_arm(&do_trap_hook_rec)))
		goto err;
	if ((retval = hook_exit_arm(&do_pgck_hook_rec)))
		goto err1;
	if ((retval = hook_exit_arm(&do_iabr_hook_rec)))
		goto err2;
	if ((retval = hook_exit_arm(&do_debug_hook_rec)))
		goto err3;
	if ((retval = hook_exit_arm(&do_page_fault_hook_rec)))
		goto err4;
	return 0;

err4:	hook_exit_disarm(&do_debug_hook_rec);
err3:	hook_exit_disarm(&do_iabr_hook_rec);
err2:	hook_exit_disarm(&do_pgck_hook_rec);
err1:	hook_exit_disarm(&do_trap_hook_rec);
err:	return retval;
}

int disarm_all_asm_dprobes_hooks(void)
{
	hook_exit_disarm(&do_page_fault_hook_rec);
	hook_exit_disarm(&do_debug_hook_rec);
	hook_exit_disarm(&do_iabr_hook_rec);
	hook_exit_disarm(&do_pgck_hook_rec);
	hook_exit_disarm(&do_trap_hook_rec);
	return 0;
}

void dprobes_hooks_asm_code_end(void) { }
