/*
 * Handle mapping of the flash on the VESUVIUS
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/concat.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define MAX_NUM_FLASH_BANKS 2
static struct mtd_info *mymtd[MAX_NUM_FLASH_BANKS] = { NULL, NULL };
static struct mtd_info *concat =  NULL;

/* Main Flash - Micron 28F128J3, 16MB (64 uniform 128K sectors)
 * This device will ALWAYS exist
 */
static struct mtd_info *main_mtd;

#define MAIN_FLASH_ADDR 0xFF000000
#define MAIN_FLASH_SIZE 0x01000000

static struct map_info vesuvius_main_map = {
	.name = "VESUVIUS_MAIN_FLASH",
	.size = MAIN_FLASH_SIZE,
	.buswidth = 1,
	.phys = MAIN_FLASH_ADDR,
};

/* FF000000 - FF01FFFF is boot strap vector - Don't mess with this!!!! */

#define APP_PARTITION_OFFSET        (0x00020000) /* FF000000 - FFBFFFFF */
#define APP_PARTITION_SIZE          (0x00CE0000) /* 12MB - 128K */

#define USYSTEM_PARTITION_OFFSET    (0x00D00000) /* FFC00000 - FFEFFFFF */
#define USYSTEM_PARTITION_SIZE      (0x00200000) /* 3MB */

#define UBOOT_PARTITION_OFFSET      (0x00F00000) /* FFF00000 - FFF3FFFF */
#define UBOOT_PARTITION_SIZE        (0x00040000) /* 256K */

/* FFF40000 - FFF60000 is UBoot Config Data */
#define UBOOT_CFG_PARTITION_OFFSET  (0x00F40000) /* FFF40000 - FFFFFFFF */
#define UBOOT_CFG_PARTITION_SIZE    (0x00020000) /* 128K */ 

#define VPD_PARTITION_OFFSET        (0x00F60000) /* FFF40000 - FFFFFFFF */
#define VPD_PARTITION_SIZE          (0x00020000) /* 128K */ 

#define EXTRA_PARTITION_OFFSET      (0x00F80000) /* FFF40000 - FFFFFFFF */
#define EXTRA_PARTITION_SIZE        (0x00080000) /* 128K */ 

/*
 * MTD 'PARTITIONING' STUFF 
 */
static struct mtd_partition main_partitions[] = {
    { .name =   "VESUVIUS Applications", 
      .offset = APP_PARTITION_OFFSET, 
      .size =   APP_PARTITION_SIZE            	      },
    { .name =   "VESUVIUS Kernel/Root File System", 
      .offset = USYSTEM_PARTITION_OFFSET, 
      .size =   USYSTEM_PARTITION_SIZE           },
    { .name =   "VESUVIUS Bootloader", 
      .offset = UBOOT_PARTITION_OFFSET, 
      .size =   UBOOT_PARTITION_SIZE            },
    { .name =   "VESUVIUS Bootloader Environment", 
      .offset = UBOOT_CFG_PARTITION_OFFSET, 
      .size =   UBOOT_CFG_PARTITION_SIZE        },
    { .name =   "VESUVIUS VPD", 
      .offset = VPD_PARTITION_OFFSET, 
      .size =   VPD_PARTITION_SIZE             },
    { .name =   "VESUVIUS Extra", 
      .offset = EXTRA_PARTITION_OFFSET, 
      .size =   EXTRA_PARTITION_SIZE             },
};

#define NUM_MAIN_PARTITIONS         (sizeof(main_partitions)/sizeof(struct mtd_partition))

/* Aux Flash - Micron 28F128J3, 16MB (64 uniform 128K sectors)
 * This device may not exist
 */
static struct mtd_info *aux_mtd;

#define AUX_FLASH_ADDR 0xFE000000
#define AUX_FLASH_SIZE 0x01000000

static struct map_info vesuvius_aux_map = {
	.name = "VESUVIUS_AUX_FLASH",
	.size = AUX_FLASH_SIZE,
	.buswidth = 1,
	.phys = AUX_FLASH_ADDR,
};

#define AUX_PARTITION_OFFSET        (0x00020000) /* FE000000 - FEFFFFFF */
#define AUX_PARTITION_SIZE          (0x01000000) /* 16 MB */

/*
 * MTD 'PARTITIONING' STUFF 
 */
static struct mtd_partition aux_partitions[] =  {
    { .name =   "VESUVIUS Auxiliary", 
      .offset = AUX_PARTITION_OFFSET, 
      .size =   AUX_PARTITION_SIZE              },
};

#define NUM_AUX_PARTITIONS         (sizeof(aux_partitions)/sizeof(struct mtd_partition))

int __init init_vesuvius(void)
{
	int retval = 0;

	printk(KERN_NOTICE "VESUVIUS main flash device: %x at %x\n", MAIN_FLASH_SIZE, MAIN_FLASH_ADDR);

	vesuvius_main_map.virt = (unsigned long)ioremap(MAIN_FLASH_ADDR, MAIN_FLASH_SIZE);

	if (!vesuvius_main_map.virt) {
		printk("Failed to ioremap\n");
		return -EIO;
	}

	simple_map_init(&vesuvius_main_map);

	main_mtd = do_map_probe("cfi_probe", &vesuvius_main_map);

        mymtd[1] = main_mtd;

	if (main_mtd) {

		main_mtd->owner = THIS_MODULE;


	        vesuvius_aux_map.virt = (unsigned long)ioremap(AUX_FLASH_ADDR, AUX_FLASH_SIZE);

        	if (!vesuvius_aux_map.virt) {
		        printk("Failed to ioremap\n");
		        return -EIO;
	        }

        	simple_map_init(&vesuvius_aux_map);

        	aux_mtd = do_map_probe("cfi_probe", &vesuvius_aux_map);
               
                mymtd[0] = aux_mtd;

                if (!aux_mtd)
       			return -ENXIO;

		aux_mtd->owner = THIS_MODULE;

        }
 
        concat = mtd_concat_create(mymtd,MAX_NUM_FLASH_BANKS,"Vesuvius Flash");               
       
        if (concat) {
                int i;

		main_partitions[0].size += aux_mtd->size;
	 
                for (i=1; i<NUM_MAIN_PARTITIONS; i++)
                  main_partitions[i].offset += aux_mtd->size;

                add_mtd_partitions(concat,main_partitions, NUM_MAIN_PARTITIONS);  
       
                return 0;
        }
       	return -ENXIO;

#if 0
		retval = add_mtd_partitions(main_mtd, main_partitions, NUM_MAIN_PARTITIONS );
        
        if (!retval) {
	        printk(KERN_NOTICE "VESUVIUS aux flash device: %x at %x\n", AUX_FLASH_SIZE, AUX_FLASH_ADDR);

	        vesuvius_aux_map.virt = (unsigned long)ioremap(AUX_FLASH_ADDR, AUX_FLASH_SIZE);

        	if (!vesuvius_aux_map.virt) {
		        printk("Failed to ioremap\n");
		        return -EIO;
	        }


        	if (aux_mtd) {
		        aux_mtd->owner = THIS_MODULE;
        		add_mtd_partitions(aux_mtd, aux_partitions, NUM_AUX_PARTITIONS );
	        }
        }
        	return retval;
	}
#endif 
	return -ENXIO;
}

static void __exit cleanup_vesuvius(void)
{
	if (main_mtd) {
		del_mtd_device(main_mtd);
		map_destroy(main_mtd);
	}
	if (vesuvius_main_map.virt) {
		iounmap((void *)vesuvius_main_map.virt);
		vesuvius_main_map.virt = 0;
	}
	if (aux_mtd) {
		del_mtd_device(aux_mtd);
		map_destroy(aux_mtd);
	}
	if (vesuvius_aux_map.virt) {
		iounmap((void *)vesuvius_aux_map.virt);
		vesuvius_aux_map.virt = 0;
	}
}

module_init(init_vesuvius);
module_exit(cleanup_vesuvius);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("LVL7 Systems.");
MODULE_DESCRIPTION("MTD map driver for VESUVIUS boards");
