/*
 *  linux/drivers/video/s3trio.c -- S3 Trio64V2 frame buffer device
 *
 *  Work in progress, i.e. for now it is just an ugly hack.
 *
 *	Copyright (C) 2003 Jochen Roth (jochen@panix.com)
 *
 *  Development of this driver was supported by funding from the
 *  City Council of La Corua, Spain.
 *
 *  This driver is based on the cyberfb.c driver, with parts from
 *  virgefb.c and radeonfb.c added. The following notices come
 *  from these drivers:
 *
 *	Copyright (C) 1997 Peter De Schrijver
 *      Copyright (C) 1997 Andr Heynatz
 *      Copyright (C) 1996 Martin Apel
 *	Copyright (C) 1996 Paul Mackerras
 *	Copyright (C) 1997 Geert Uytterhoeven
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/selection.h>
#include <asm/io.h>
#include <asm/prom.h>
#include <linux/pci.h>
#include <asm/pci-bridge.h>

#ifdef CONFIG_FB_COMPAT_XPMAC
#include <asm/vc_ioctl.h>
#endif

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#include "s3trio.h"


/* Debug diagnostics */
#define	TRIO_DUMP_REGS		0
#define	TRIO_DUMP_PCICFG	1

#define	debug_msg	if(1) printk
#define	debug_flow	if(0) printk
#define	debug_init	if(1) printk
#define	debug_reg	if(0) printk



/* frame buffer context structure */

typedef struct triofb {
	struct fb_info fb_info;

	int accel;
	struct pci_dev *pdev;
	struct display disp;
	int currcon;
	struct display *currcon_display;
	struct { u_char red, green, blue, pad; } palette[256];
	char name[16];
	uint32_t bar0_phys;
	uint32_t bar0_size;
	char *bar0_virt;
	uint32_t mmio_phys;
	uint32_t mmio_size;
	char *mmio_virt;
	char *mmio_regs;
	uint32_t vram_phys;
	uint32_t vram_size;
	char *vram_virt;
	long memclk;

	uint32_t bytes_per_pixel;

#if defined(FBCON_HAS_CFB16) || defined(FBCON_HAS_CFB24) || defined(FBCON_HAS_CFB32)
	union {
#ifdef FBCON_HAS_CFB16
                u_int16_t       cfb16[16];
#endif
#ifdef FBCON_HAS_CFB24
                u_int32_t       cfb24[16];
#endif
#ifdef FBCON_HAS_CFB32
                u_int32_t       cfb32[16];
#endif
	} con_cmap;
#endif /* FBCON_HAS_CFB16 | FBCON_HAS_CFB24 | FBCON_HAS_CFB32 */
} triofb_t;


/* internal functions */

static int __init s3triofb_pci_init(struct pci_dev *pdev);

static void Trio_WaitQueue(triofb_t *trio, u_short fifo);
static void Trio_WaitBlit(triofb_t *trio);
static void Trio_BitBLT(triofb_t *trio, u_short curx, u_short cury, u_short destx,
		u_short desty, u_short width, u_short height, u_short mode);
static void Trio_RectFill(triofb_t *trio, u_short x, u_short y, u_short width, u_short height,
		u_short mode, u_short color);
static void Trio_MoveCursor(triofb_t *trio, u_short x, u_short y);

static int s3trio_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
		u_int *transp, struct fb_info *info);
static int s3trio_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
		u_int transp, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);


/* frame buffer device interface */

static int s3triofb_op_get_fix(struct fb_fix_screeninfo *fix, int con,
			  struct fb_info *info);
static int s3triofb_op_get_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info);
static int s3triofb_op_set_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info);
static int s3triofb_op_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info);
static int s3triofb_op_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info);
static int s3triofb_op_pan_display(struct fb_var_screeninfo *var, int con,
			      struct fb_info *info);

static struct fb_ops s3trio_ops = {
	owner:		THIS_MODULE,
	fb_get_fix:	s3triofb_op_get_fix,
	fb_get_var:	s3triofb_op_get_var,
	fb_set_var:	s3triofb_op_set_var,
	fb_get_cmap:	s3triofb_op_get_cmap,
	fb_set_cmap:	s3triofb_op_set_cmap,
	fb_pan_display:	s3triofb_op_pan_display,
};


/* frame buffer console device interface */

static int s3triofbcon_switch(int con, struct fb_info *info);
static int s3triofbcon_updatevar(int con, struct fb_info *info);
static void s3triofbcon_blank(int blank, struct fb_info *info);
#if 0
static int s3triofbcon_setcmap(struct fb_cmap *cmap, int con);
#endif
#ifdef FBCON_HAS_CFB8
static struct display_switch s3triofb_cfb8;
#endif
#ifdef FBCON_HAS_CFB16
static struct display_switch s3triofb_cfb16;
#endif
#ifdef FBCON_HAS_CFB32
static struct display_switch s3triofb_cfb32;
#endif



/* pci device driver linkage */

static struct pci_device_id s3triofb_pci_table[] __devinitdata = {
	{ PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_TRIO64V2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{ 0, }
};

MODULE_DEVICE_TABLE(pci, s3triofb_pci_table);

static int s3triofb_pci_register(struct pci_dev *pdev, const struct pci_device_id *ent);
static void __devexit s3triofb_pci_unregister(struct pci_dev *pdev);

static struct pci_driver s3triofb_driver = {
	name:		"s3triofb",
	id_table:	s3triofb_pci_table,
	probe:		s3triofb_pci_register,
	remove:		__devexit_p(s3triofb_pci_unregister),
};


/* module linkage */

int __init s3triofb_init(void);
void __exit s3triofb_exit(void);
int __init s3triofb_setup(char *options);
#ifdef MODULE
module_init(s3triofb_init);
module_exit(s3triofb_exit);
#endif

MODULE_AUTHOR("Jochen Roth");
MODULE_DESCRIPTION("framebuffer driver for S3Trio64V2");
MODULE_LICENSE("GPL");



/* local data */

static triofb_t xox;

static struct fb_var_screeninfo s3triofb_default_var = {
		1024, 768, 1024, 768, 0, 0, 8, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCELF_TEXT, 15385, 160, 24, 29, 3, 136, 6,
#if 0
		0, 0, -1, -1, FB_ACCELF_TEXT, 16667, 224, 72, 60, 12, 168, 4,
#endif
		0,
		FB_VMODE_NONINTERLACED
};

static struct fb_var_screeninfo s3triofb_default_var2 = {
		/* xres yres xres_virtual yres_virtual xoffset yoffset bits_per_pixel grayscale */
		640, 480, 640, 480, 0, 0, 8, 0,
		/* rgbt bitfield def */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		/* nonstd activate height width accel_flags */
		0, 0, -1, -1, FB_ACCELF_TEXT,
		/* pixclock (ps) left_margin right_margin upper_margin lower_margin hsync_len vsync_len */
		39722, 40, 24, 32, 11, 96, 2,
		/* sync vmode */
		0, FB_VMODE_NONINTERLACED
};

static struct fb_var_screeninfo s3triofb_default_var3 = {
		800, 600, 800, 600, 0, 0, 8, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCELF_TEXT, 25000, 88, 40, 23, 1, 128, 4,
		FB_SYNC_HOR_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT,
		FB_VMODE_NONINTERLACED
};

#if 0
		640, 480, 640, 480, 0, 0, 8, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCELF_TEXT, 31250, 160, 136, 82, 61, 88, 2,
		0, FB_VMODE_NONINTERLACED

		0, 0, -1, -1, FB_ACCELF_TEXT, 27778, 64, 24, 22, 1, 72, 2,

		640, 480, 640, 480, 0, 0, 8, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCELF_TEXT, 39722, 40, 24, 32, 11, 96, 2,
		FB_SYNC_COMP_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT, 
		FB_VMODE_NONINTERLACED

		640, 480, 640, 480, 0, 0, 8, 0,
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		0, 0, -1, -1, FB_ACCELF_TEXT, 31250, 160, 136, 82, 61, 88, 2,
		0, FB_VMODE_NONINTERLACED

static struct fb_var_screeninfo default_var_640x480 = {
		/* xres yres xres_virtual yres_virtual xoffset yoffset bits_per_pixel grayscale */
		640, 480, 640, 480, 0, 0, 8, 0,
		/* rgbt bitfield def */
		{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
		/* nonstd activate height width accel_flags */
		0, 0, -1, -1, FB_ACCELF_TEXT,
		/* pixclock (ps) left_margin rigth_margin upper_margin lower_margin hsync_len vsync_len */
		39722, 40, 24, 32, 11, 96, 2,
		/* sync vmode */
		FB_SYNC_COMP_HIGH_ACT|FB_SYNC_VERT_HIGH_ACT, FB_VMODE_NONINTERLACED
};
#endif



/* io-mapped and memory-mapped register access wrappers */

static void s3_mmio_put8(triofb_t *trio, uint32_t addr, uint8_t value, char *a2)
{
	if((addr & ~1) != 0x3c8)
	debug_reg("mmio_put8  0x%04x %02x\n", addr, value);
	if((addr & 0xfff) != ((uint32_t)a2 & 0xfff)) printk("AAAAAAAAAAAAAAAAAAAH: s3_mmio_put8 %x %x\n", addr, (uint32_t)a2);
	out_8((void*)(trio->mmio_regs + addr), value);
}

static uint8_t s3_mmio_get8(triofb_t *trio, uint32_t addr)
{
	return(in_8((void *)(trio->mmio_regs + addr)));
}

static void s3_mmio_put16(triofb_t *trio, uint32_t addr, uint16_t value)
{
	debug_reg("mmio_put16 0x%04x %02x\n", addr, value);
	out_le16((void*)(trio->mmio_regs + addr), value);
}

static uint16_t s3_mmio_get16(triofb_t *trio, uint32_t addr)
{
	return(in_le16((void *)(trio->mmio_regs + addr)));
}

static void s3_outb(triofb_t *trio, uint32_t port, uint8_t value)
{
	debug_reg("mmio_outb  0x%04x %02x\n", port, value);
	outb(value, port);
}

static uint8_t s3_inb(triofb_t *trio, uint32_t port)
{
	return(inb(port));
}

static void s3_outw(triofb_t *trio, uint32_t port, uint16_t value, uint32_t p2)
{
	debug_reg("mmio_outw  0x%04x %02x\n", port, value);
	if(port != p2) printk("AAAAAAAAAAAAAAAAAAAH: s3_outw %x %x\n", port, p2);
	outw(value, port);
}

static uint16_t s3_inw(triofb_t *trio, uint32_t port)
{
	return(inw(port));
}


static void s3_seq_outb(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("seq_outb   0x%04x %02x\n", index, value);
	outb(index, 0x3c4);
	outb(value, 0x3c5);
}

static uint8_t s3_seq_inb(triofb_t *trio, int index)
{
	outb(index, 0x3c4);
	return(inb(0x3c5));
}

static void s3_crtc_outb(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("crtc_outb  0x%04x %02x\n", index, value);
	outb(index, 0x3d4);
	outb(value, 0x3d5);
}

static uint8_t s3_crtc_inb(triofb_t *trio, int index)
{
	outb(index, 0x3d4);
	return(inb(0x3d4+1));
}

static void s3_seq_write(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("seq_write  0x%04x %02x\n", index, value);
	out_8(trio->mmio_regs + 0x3c4, index);
	out_8(trio->mmio_regs + 0x3c5, value);
}

static uint8_t s3_seq_read(triofb_t *trio, int index)
{
	out_8(trio->mmio_regs + 0x3c4, index);
	return(in_8(trio->mmio_regs + 0x3c5));
}


static void s3_crtc_write(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("crtc_write 0x%04x %02x\n", index, value);
	out_8(trio->mmio_regs + 0x3d4, index);
	out_8(trio->mmio_regs + 0x3d5, value);
}

static uint8_t s3_crtc_read(triofb_t *trio, int index)
{
	out_8(trio->mmio_regs + 0x3d4, index);
	return(in_8(trio->mmio_regs + 0x3d5));
}

static void s3_gc_write(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("gc_write   0x%04x %02x\n", index, value);
	out_8(trio->mmio_regs + 0x3ce, index);
	out_8(trio->mmio_regs + 0x3cf, value);
}

static uint8_t s3_gc_read(triofb_t *trio, int index)
{
	out_8(trio->mmio_regs + 0x3ce, index);
	return(in_8(trio->mmio_regs + 0x3cf));
}

static void s3_ac_outb(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("ac_outb    0x%04x %02x\n", index, value);
	inb(0x3da);
	outb(index, 0x3c0);
	if(index < 0x20)
		outb(value, 0x3c0);
}

static void s3_ac_write(triofb_t *trio, int index, uint8_t value)
{
	debug_reg("ac_write   0x%04x %02x\n", index, value);
	in_8(trio->mmio_regs + 0x3da);
	out_8(trio->mmio_regs + 0x3c0, index);
	if(index < 0x20) {
		out_8(trio->mmio_regs + 0x3c0, value);
		out_8(trio->mmio_regs + 0x3c0, 0x20);
	}
}

static uint8_t s3_ac_read(triofb_t *trio, int index)
{
	uint8_t value;
	in_8(trio->mmio_regs + 0x3da);
	out_8(trio->mmio_regs + 0x3c0, index);
	value = in_8(trio->mmio_regs + 0x3c1);
	in_8(trio->mmio_regs + 0x3da);
	out_8(trio->mmio_regs + 0x3c0, 0x20);
	return(value);
}


#if TRIO_DUMP_PCICFG
static void trio_dump_pcicfg(struct pci_dev *pdev)
{
	uint32_t t32;

	pci_read_config_dword(pdev, PCI_VENDOR_ID, &t32);
	printk("s3triofb: bus=0x%x devfn=0x%x id=0x%x\n", pdev->bus->number, pdev->devfn, t32);
	pci_read_config_dword(pdev, 4, &t32);
	printk("s3triofb: cmdst = 0x%x\n", t32);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, &t32);
	printk("s3triofb: bar0 = 0x%x\n", t32);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_1, &t32);
	printk("s3triofb: bar1 = 0x%x\n", t32);
	printk("s3triofb: rsrc0 = 0x%lx, 0x%lx\n", pci_resource_start(pdev, 0), pci_resource_len(pdev, 0));
}
#endif

/* dump chip registers */

#if TRIO_DUMP_REGS
static uint trio_pll_clock(uint lo, uint hi)
{
	uint clk, clk_shift;
	clk = (0x37EE * (hi + 2)) / ((lo & 0x1F) + 2);
	clk_shift = (lo >> 5) & 0x03;
	if(clk_shift) {
		clk = clk >> clk_shift;
	}
	return(clk * 0x3E8);
}

static void trio_dump_regs(triofb_t *trio, char *msg)
{
	volatile uint8_t *regs = trio->mmio_regs, index;
	int cols;
	static int crtc_regs[] = {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
		0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
		0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
		0x18,
		            0x22,       0x24,       0x26,
		                              0x2d, 0x2e, 0x2f,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
		0x38, 0x39, 0x3a, 0x3b, 0x3c,
		0x40, 0x41, 0x42, 0x43,       0x45, 0x46, 0x47,
		0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
		0x58, 0x59, 0x5a,       0x5c, 0x5d, 0x5e,
		0x60, 0x61,       0x63,       0x65, 0x66, 0x67,
		0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
		0x70, 0x71, 0x72,
		0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
		0x88,
		0x90, 0x91, 0x92, 0x93,
		0xff };
#if 0
	static int seq_regs[] = {
		0x01, 0x02, 0x03, 0x04, 0x08, 0x09, 0x0a, 0x0b, 0x0d,
		0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
		0x18, 0xff };
#endif

	if(msg) printk("trio_dump_regs %s\n", msg);

	printk("MISC=0x%02x\n", *(regs + GREG_MISC_OUTPUT_R));

        /* Dump the VGA setup values */
	for(index = cols = 0; crtc_regs[index] != 0xff; index++) {
		*(regs + S3_CRTC_ADDR) = crtc_regs[index];
		printk("  CR%02x=0x%02x", crtc_regs[index], *(regs + S3_CRTC_DATA));
		if(++cols == 6) { printk("\n"); cols = 0; }
	}
	if(cols) printk("\n");

#if 0
	for(index = cols = 0; seq_regs[index] != 0xff; index++) {
		*(regs + SEQ_INDEX) = seq_regs[index];
		printk("SEQ%02x=0x%02x  ", seq_regs[index], *(regs + SEQ_DATA));
#endif
	for(index = cols = 0; index <= 0x1c; index++) {
		*(regs + SEQ_INDEX) = index;
		printk("  SR%02x=0x%02x", index, *(regs + SEQ_DATA));
		if(++cols == 6) { printk("\n"); cols = 0; }
	}
	if(cols) printk("\n");

	for(index = cols = 0; index <= 0x08; index++) {
		printk("  GR%02x=0x%02x", index, s3_gc_read(trio, index));
		if(++cols == 6) { printk("\n"); cols = 0; }
	}
	if(cols) printk("\n");

	for(index = cols = 0; index <= 0x14; index++) {
		printk("  AR%02x=0x%02x", index, s3_ac_read(trio, index));
		if(++cols == 6) { printk("\n"); cols = 0; }
	}
	if(cols) printk("\n");

	printk("dclk = %d   mclk = %d\n",
		trio_pll_clock(s3_seq_read(trio, 0x12), s3_seq_read(trio, 0x13)),
		trio_pll_clock(s3_seq_read(trio, 0x10), s3_seq_read(trio, 0x11)));

	if(msg) printk("trio_dump_regs %s end\n\n", msg);

	return;
}
#endif


static int s3_unlock(triofb_t *trio)
{
	uint8_t cr38, cr39, cr40;

	debug_init("s3_unlock()\n");

	/* unlock s3 */
	s3_outb(trio, 0x3c3, 0x01);

	/* Make sure chip is in color emulation mode */
	s3_outb(trio, 0x3c2, s3_inb(trio, 0x03cc) | 1);

	/* Write code to CR38 to provide access to extended
	   CRTC registers CR2D-CR3F */
	cr38 = s3_crtc_inb(trio, 0x38);
	s3_crtc_outb(trio, 0x38, 0x48);

	/* Write code to CR39 to provide access to extended
	   CRTC registers CR40-CRFF */
	cr39 = s3_crtc_inb(trio, 0x39);
	s3_crtc_outb(trio, 0x39, 0xA0);

	debug_init("s3_unlock: cr33=%02x cr38=%02x cr39=%02x cr40=%02x sr08=%02x\n",
		s3_crtc_inb(trio, 0x33),
		cr38, cr39,
		s3_crtc_inb(trio, 0x40),
		s3_seq_inb(trio, 0x08));
		

	s3_crtc_outb(trio, 0x33, 0x02);

	/* ??? Set bit 0 in CR40 to enable access to the Enhanced Programming registers */
#if 0
	outb(0x33,0x3d4);
	outw(IO_OUT16VAL((inb(0x3d5) & ~(0x2 | 0x10 |  0x40)) | 0x20, 0x33), 0x3d4);
#endif
#if 0
	cr40 = s3_crtc_inb(trio, 0x33);
	s3_crtc_outb(trio, 0x33, (cr40 & ~(0x2 | 0x10 | 0x40)) | 0x20);
#endif
	cr40 = s3_crtc_inb(trio, 0x40);
	s3_crtc_outb(trio, 0x40, cr40 | 0x01);

	/* Loading xxxx0110b unlocks extended sequencer registers. */
	s3_seq_outb(trio, 0x08, 0x06);

	return(0);
}

static int s3_chip_type(triofb_t *trio)
{
	uint8_t r2d, r2e, r2f;

	r2d = s3_crtc_inb(trio, 0x2d);
	r2e = s3_crtc_inb(trio, 0x2e);
	r2f = 0;
	if(r2d == 0x88 && r2e == 0x11) {
		/* trio */
		r2f = s3_crtc_inb(trio, 0x2f);
	}
	debug_init("s3_chip_type: r2d=0x%02x r2e=0x%02x r2f=0x%02x\n", r2d, r2e, r2f);

	return(0);
}



static void s3trio_resetaccel(triofb_t *trio)
{

/* subsystem control register 18-2 */
#define	EC00_REG	0x42e8
#define EC00_RESET      0x8000
#define EC00_ENABLE     0x4000

/* advanced function control register 18-3 */
#define	EC01_REG	0x4ae8
#define EC01_ENH_ENB    0x0005
#define EC01_LAW_ENB    0x0010
#define EC01_MMIO_ENB   0x0020

#define MF_MULT_MISC    0xE000

#define SRC_FOREGROUND  0x0020
#define SRC_BACKGROUND  0x0000
#define MIX_SRC                 0x0007
#define MF_T_CLIP       0x1000
#define MF_L_CLIP       0x2000
#define MF_B_CLIP       0x3000
#define MF_R_CLIP       0x4000
#define MF_PIX_CONTROL  0xA000
#define MFA_SRC_FOREGR_MIX      0x0000
#define MF_PIX_CONTROL  0xA000

	s3_outw(trio, 0x42e8, EC00_RESET, 0x42e8);
	s3_inw(trio, 0x42e8);
	s3_outw(trio, 0x42e8, EC00_ENABLE, 0x42e8);
	s3_inw(trio, 0x42e8);
	s3_outw(trio, 0x4ae8, EC01_ENH_ENB | EC01_LAW_ENB, 0x4ae8);
	s3_outw(trio, 0xbee8, MF_MULT_MISC, 0xbee8);	/* 16 bit I/O registers */

	/* Now set some basic accelerator registers */
	Trio_WaitQueue(trio, 0x0400);
	s3_outw(trio, 0xbae8, SRC_FOREGROUND | MIX_SRC, 0xbae8);
	s3_outw(trio, 0xb6e8, SRC_BACKGROUND | MIX_SRC, 0xb6e8);	/* direct color*/
	s3_outw(trio, 0xbee8, MF_T_CLIP | 0, 0xbee8 );	/* clip virtual area  */
	s3_outw(trio, 0xbee8, MF_L_CLIP | 0, 0xbee8 );
	s3_outw(trio, 0xbee8, MF_R_CLIP | (trio->disp.var.xres_virtual - 1), 0xbee8);
	s3_outw(trio, 0xbee8, MF_B_CLIP | (trio->disp.var.yres_virtual - 1), 0xbee8);
	Trio_WaitQueue(trio, 0x0400);
	s3_outw(trio, 0xaae8, 0xffff, 0xaae8);		/* Enable all planes */
	s3_outw(trio, 0xaae8, 0xffff, 0xaae8);		/* Enable all planes */
	s3_outw(trio, 0xbee8, MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, 0xbee8);
}






static int s3_setup(triofb_t *trio)
{
	u_char temp_uc;
	int i;
	u_long *CursorBase;

	debug_flow("s3_setup enter\n");
	s3_unlock(trio);

	s3_chip_type(trio);

	/* linear address window control register */

	temp_uc = s3_crtc_inb(trio, S3_CRT_LAWC);
	s3_crtc_outb(trio, S3_CRT_LAWC, temp_uc | 0x13);
	s3_crtc_outb(trio, S3_CRT_EMC1, 8);

	/* not: switch off I/O accesses */

	s3trio_resetaccel(trio);

	s3_mmio_put8(trio, 0x3d4, 0x30, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d4, 0x2d, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d4, 0x2e, trio->mmio_regs + 0x03D4);

	s3_mmio_put8(trio, 0x3d4, 0x50, trio->mmio_regs + 0x03D4);

	/* unlock extended crtc */
	s3_crtc_write(trio, 0x39, 0xa0);

	/* disable hw cursor and set start x/y to 0 */
	s3_crtc_write(trio, 0x45, 0);
	s3_crtc_write(trio, 0x4e, 0);
	s3_crtc_write(trio, 0x4f, 0);

#if 0	/* jr */
	/* disable HW cursor */

	s3_mmio_put8(trio, 0x3d4, 0x39, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0xa0, trio->mmio_regs + 0x03D5);

	s3_mmio_put8(trio, 0x3d4, 0x45, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0, trio->mmio_regs + 0x03D5);

	s3_mmio_put8(trio, 0x3d4, 0x4e, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0, trio->mmio_regs + 0x03D5);

	s3_mmio_put8(trio, 0x3d4, 0x4f, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0, trio->mmio_regs + 0x03D5);
#endif	/* jr */

	/* init HW cursor */

	CursorBase = (u_long *)(trio->vram_virt + 2*1024*1024 - 0x400);
	for (i = 0; i < 8; i++) {
		*(CursorBase  +(i*4)) = 0xffffff00;
		*(CursorBase+1+(i*4)) = 0xffff0000;
		*(CursorBase+2+(i*4)) = 0xffff0000;
		*(CursorBase+3+(i*4)) = 0xffff0000;
	}
	for (i = 8; i < 64; i++) {
		*(CursorBase  +(i*4)) = 0xffff0000;
		*(CursorBase+1+(i*4)) = 0xffff0000;
		*(CursorBase+2+(i*4)) = 0xffff0000;
		*(CursorBase+3+(i*4)) = 0xffff0000;
	}

	s3_crtc_write(trio, 0x4c, ((2*1024-1)&0x0f00)>>8);
	s3_crtc_write(trio, 0x4d, (2*1024-1)&0xff);

#if 0	/* jr */
	s3_mmio_put8(trio, 0x3d4, 0x4c, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, ((2*1024 - 1)&0xf00)>>8, trio->mmio_regs + 0x03D5);

	s3_mmio_put8(trio, 0x3d4, 0x4d, trio->mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, (2*1024 - 1) & 0xff, trio->mmio_regs + 0x03D5);

	s3_mmio_put8(trio, 0x3d4, 0x45, trio->mmio_regs + 0x03D4);
	s3_mmio_get8(trio, 0x3d4);
#endif	/* jr */

	s3_crtc_read(trio, 0x45);
	s3_crtc_write(trio, 0x4a, 0x80);
	s3_crtc_write(trio, 0x4a, 0x80);
	s3_crtc_write(trio, 0x4a, 0x80);
	s3_crtc_write(trio, 0x4b, 0x00);
	s3_crtc_write(trio, 0x4b, 0x00);
	s3_crtc_write(trio, 0x4b, 0x00);
	s3_crtc_write(trio, 0x45, 0);

#if 0	/* jr */
	s3_mmio_put8(trio, 0x3d4, 0x4a, mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0x80, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d5, 0x80, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d5, 0x80, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d4, 0x4b, mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0x00, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d5, 0x00, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d5, 0x00, mmio_regs + 0x03D5);
	s3_mmio_put8(trio, 0x3d4, 0x45, mmio_regs + 0x03D4);
	s3_mmio_put8(trio, 0x3d5, 0x00, mmio_regs + 0x03D5);
#endif	/* jr */


	debug_flow("s3_setup exit\n");
	return(0);
}


static void s3_display_enable(triofb_t *trio, int enable)
{
	uint8_t v;
	debug_flow("s3_display_enable %d\n", enable);
	if(enable) {
		s3_crtc_outb(trio, 0x0d, s3_crtc_inb(trio, 0x0d) & 0x0f);
	}
	/* SR1 is clock mode */
	v = s3_seq_inb(trio, 0x01) & 0xdf;
	if(!enable)
		v |= 0x20;
	s3_seq_outb(trio, 0x01, v);
}


static const unsigned char clocks[] =
{
	0x13, 0x61, 0x6b, 0x6d, 0x51, 0x69, 0x54, 0x69,
	0x4f, 0x68, 0x6b, 0x6b, 0x18, 0x61, 0x7b, 0x6c,
	0x51, 0x67, 0x24, 0x62, 0x56, 0x67, 0x77, 0x6a,
	0x1d, 0x61, 0x53, 0x66, 0x6b, 0x68, 0x79, 0x69,
	0x7c, 0x69, 0x7f, 0x69, 0x22, 0x61, 0x54, 0x65,
	0x56, 0x65, 0x58, 0x65, 0x67, 0x66, 0x41, 0x63,
	0x27, 0x61, 0x13, 0x41, 0x37, 0x62, 0x6b, 0x4d,
	0x23, 0x43, 0x51, 0x49, 0x79, 0x66, 0x54, 0x49,
	0x7d, 0x66, 0x34, 0x56, 0x4f, 0x63, 0x1f, 0x42,
	0x6b, 0x4b, 0x7e, 0x4d, 0x18, 0x41, 0x2a, 0x43,
	0x7b, 0x4c, 0x74, 0x4b, 0x51, 0x47, 0x65, 0x49,
	0x24, 0x42, 0x68, 0x49, 0x56, 0x47, 0x75, 0x4a,
	0x77, 0x4a, 0x31, 0x43, 0x1d, 0x41, 0x71, 0x49,
	0x53, 0x46, 0x29, 0x42, 0x6b, 0x48, 0x1f, 0x41,
	0x79, 0x49, 0x6f, 0x48, 0x7c, 0x49, 0x38, 0x43,
	0x7f, 0x49, 0x5d, 0x46, 0x22, 0x41, 0x53, 0x45,
	0x54, 0x45, 0x55, 0x45, 0x56, 0x45, 0x57, 0x45,
	0x58, 0x45, 0x25, 0x41, 0x67, 0x46, 0x5b, 0x45,
	0x41, 0x43, 0x78, 0x47, 0x27, 0x41, 0x51, 0x44,
	0x13, 0x21, 0x7d, 0x47, 0x37, 0x42, 0x71, 0x46,
	0x6b, 0x2d, 0x14, 0x21, 0x23, 0x23, 0x7d, 0x2f,
	0x51, 0x29, 0x61, 0x2b, 0x79, 0x46, 0x1d, 0x22,
	0x54, 0x29, 0x45, 0x27, 0x7d, 0x46, 0x7f, 0x46,
	0x4f, 0x43, 0x2f, 0x41, 0x1f, 0x22, 0x6a, 0x2b,
	0x6b, 0x2b, 0x5b, 0x29, 0x7e, 0x2d, 0x65, 0x44,
	0x18, 0x21, 0x5e, 0x29, 0x2a, 0x23, 0x45, 0x26,
	0x7b, 0x2c, 0x19, 0x21, 0x74, 0x2b, 0x75, 0x2b,
	0x51, 0x27, 0x3f, 0x25, 0x65, 0x29, 0x40, 0x25,
	0x24, 0x22, 0x41, 0x25, 0x68, 0x29, 0x42, 0x25,
	0x56, 0x27, 0x7e, 0x2b, 0x75, 0x2a, 0x1c, 0x21,
	0x77, 0x2a, 0x4f, 0x26, 0x31, 0x23, 0x6f, 0x29,
	0x1d, 0x21, 0x32, 0x23, 0x71, 0x29, 0x72, 0x29,
	0x53, 0x26, 0x69, 0x28, 0x29, 0x22, 0x75, 0x29,
	0x6b, 0x28, 0x1f, 0x21, 0x1f, 0x21, 0x6d, 0x28,
	0x79, 0x29, 0x2b, 0x22, 0x6f, 0x28, 0x59, 0x26,
	0x7c, 0x29, 0x7d, 0x29, 0x38, 0x23, 0x21, 0x21,
	0x7f, 0x29, 0x39, 0x23, 0x5d, 0x26, 0x75, 0x28,
	0x22, 0x21, 0x77, 0x28, 0x53, 0x25, 0x6c, 0x27,
	0x54, 0x25, 0x61, 0x26, 0x55, 0x25, 0x30, 0x22,
	0x56, 0x25, 0x63, 0x26, 0x57, 0x25, 0x71, 0x27,
	0x58, 0x25, 0x7f, 0x28, 0x25, 0x21, 0x74, 0x27,
	0x67, 0x26, 0x40, 0x23, 0x5b, 0x25, 0x26, 0x21,
	0x41, 0x23, 0x34, 0x22, 0x78, 0x27, 0x6b, 0x26,
	0x27, 0x21, 0x35, 0x22, 0x51, 0x24, 0x7b, 0x27,
	0x13, 0x01, 0x13, 0x01, 0x7d, 0x27, 0x4c, 0x09,
	0x37, 0x22, 0x5b, 0x0b, 0x71, 0x26, 0x5c, 0x0b,
	0x6b, 0x0d, 0x47, 0x23, 0x14, 0x01, 0x4f, 0x09,
	0x23, 0x03, 0x75, 0x26, 0x7d, 0x0f, 0x1c, 0x02,
	0x51, 0x09, 0x59, 0x24, 0x61, 0x0b, 0x69, 0x25,
	0x79, 0x26, 0x34, 0x05, 0x1d, 0x02, 0x6b, 0x25,
	0x54, 0x09, 0x35, 0x05, 0x45, 0x07, 0x6d, 0x25,
	0x7d, 0x26, 0x16, 0x01, 0x7f, 0x26, 0x77, 0x0d,
	0x4f, 0x23, 0x78, 0x0d, 0x2f, 0x21, 0x27, 0x03,
	0x1f, 0x02, 0x59, 0x09, 0x6a, 0x0b, 0x73, 0x25,
	0x6b, 0x0b, 0x63, 0x24, 0x5b, 0x09, 0x20, 0x02,
	0x7e, 0x0d, 0x4b, 0x07, 0x65, 0x24, 0x43, 0x22,
	0x18, 0x01, 0x6f, 0x0b, 0x5e, 0x09, 0x70, 0x0b,
	0x2a, 0x03, 0x33, 0x04, 0x45, 0x06, 0x60, 0x09,
	0x7b, 0x0c, 0x19, 0x01, 0x19, 0x01, 0x7d, 0x0c,
	0x74, 0x0b, 0x50, 0x07, 0x75, 0x0b, 0x63, 0x09,
	0x51, 0x07, 0x23, 0x02, 0x3f, 0x05, 0x1a, 0x01,
	0x65, 0x09, 0x2d, 0x03, 0x40, 0x05, 0x00, 0x00,
};

/*
 * Computes M, N, and R values from
 * given input frequency. It uses a table of
 * precomputed values, to keep CPU time low.
 *
 * The return value consist of:
 * lower byte:  Bits 4-0: N Divider Value
 *	        Bits 5-6: R Value          for e.g. SR10 or SR12
 * higher byte: Bits 0-6: M divider value  for e.g. SR11 or SR13
 */
static uint16_t s3_compute_clock(unsigned long freq)
{
	unsigned char *mnr, *save;	/* M, N + R vals */
	unsigned long work_freq, r, actual_freq;
	unsigned short erg;
	long diff, d2;

	if (freq < 12500000 || freq > 135000000) {
		printk("s3triofb: illegal clock frequency %ld, using 25MHz\n", freq);
		freq = 25000000;
	}

	mnr = (unsigned char*)clocks;
	d2 = 0x7fffffff;

	while (*mnr) {	/* mnr vals are 0-terminated */
		work_freq = (0x37EE * (mnr[0] + 2)) / ((mnr[1] & 0x1F) + 2);

		r = (mnr[1] >> 5) & 0x03;
		if (r != 0) {
			work_freq = work_freq >> r;	/* r is the freq divider */
		}

		work_freq *= 0x3E8;	/* 2nd part of OSC */

		diff = abs(freq - work_freq);

		if (d2 >= diff) {
			actual_freq = work_freq;
			d2 = diff;
			/* In save are the vals for minimal diff */
			save = mnr;
		}
		mnr += 2;
	}
	erg = *((unsigned short *)save);

	debug_msg("s3_compute_clock(%ld): got 0x%04x %ld\n", freq, erg, actual_freq);
	return(erg);
}

#if 0
static int cv_has_4mb (volatile unsigned char *fb)
{
	volatile unsigned long *tr, *tw;
	DPRINTK("ENTER\n");

	/* write patterns in memory and test if they can be read */
	tw = (volatile unsigned long *) fb;
	tr = (volatile unsigned long *) (fb + 0x02000000);

	*tw = 0x87654321;
	
	if (*tr != 0x87654321) {
		DPRINTK("EXIT - <4MB\n");
		return (0);
	}

	/* upper memory region */
	tw = (volatile unsigned long *) (fb + 0x00200000);
	tr = (volatile unsigned long *) (fb + 0x02200000);

	*tw = 0x87654321;

	if (*tr != 0x87654321) {
		DPRINTK("EXIT - <4MB\n");
		return (0);
	}

	*tw = 0xAAAAAAAA;

	if (*tr != 0xAAAAAAAA) {
		DPRINTK("EXIT - <4MB\n");
		return (0);
	}

	*tw = 0x55555555;

	if (*tr != 0x55555555) {
		DPRINTK("EXIT - <4MB\n");
		return (0);
	}

	DPRINTK("EXIT\n");
	return (1);
}
#endif


static int s3_load_video_mode(triofb_t *trio, struct fb_var_screeninfo *video_mode)
{
	unsigned short mnr;
	int test;
	int fx, fy;
	unsigned short HT, HDE, HBS, HBE, HSS, HSE, VDE, VBS, VBE, VSS, VSE, VT;
	char isInterlaced, isDoubleScan, isTextMode, isConsole;
	int misc;
	int cr50, cr53, sr15, sr18, clock_mode;
	int m, n;
	int tfillm, temptym;
	int hmul;

	int xres, hfront, hsync, hback;
	int yres, vfront, vsync, vback;
	int bpp;
	long freq;

	debug_flow("s3_load_video_mode enter\n");

	s3_display_enable(trio, 0);

#if 0
	s3trio_resetaccel(trio);
#endif

	isTextMode = 0;		/* if depth == 4 */
	isConsole = 0;		/* mode num == 255 (console) */
	fx = fy = 8;		/* force 8x8 font */

	switch (video_mode->bits_per_pixel) {
	case 15:
	case 16:
		hmul = 2;
		break;
	default:
		hmul = 1;
		break;
	}

	bpp = video_mode->bits_per_pixel;
	xres = video_mode->xres;
	hfront = video_mode->right_margin;
	hsync = video_mode->hsync_len;
	hback = video_mode->left_margin;

	isInterlaced = 0;
	isDoubleScan = 0;

	if (video_mode->vmode & FB_VMODE_DOUBLE) {
		yres = video_mode->yres * 2;
		vfront = video_mode->lower_margin * 2;
		vsync = video_mode->vsync_len * 2;
		vback = video_mode->upper_margin * 2;
		isDoubleScan = 1;
	} else if (video_mode->vmode & FB_VMODE_INTERLACED) {
		yres = (video_mode->yres + 1) / 2;
		vfront = (video_mode->lower_margin + 1) / 2;
		vsync = (video_mode->vsync_len + 1) / 2;
		vback = (video_mode->upper_margin + 1) / 2;
		isInterlaced = 1;
	} else {
		yres = video_mode->yres;
		vfront = video_mode->lower_margin;
		vsync = video_mode->vsync_len;
		vback = video_mode->upper_margin;
	}

	HBS = hmul * (xres / 8);
	HBE = hmul * ((xres/8) + (hfront/8) + (hsync/8) + (hback/8) - 2);
	HSS = hmul * ((xres/8) + (hfront/8) + 2);
	HSE = hmul * ((xres/8) + (hfront/8) + (hsync/8) + 1);
	HT  = hmul * ((xres/8) + (hfront/8) + (hsync/8) + (hback/8));

	VBS = yres;
	VBE = yres + vfront + vsync + vback - 2;
	VSS = yres + vfront - 1;
	VSE = yres + vfront + vsync - 1;
	VT  = yres + vfront + vsync + vback - 2;

#if 0
	if (isTextMode)
		par->HorizDisplayEnd = ((video_mode->xres + fx - 1) / fx) - 1;
	else
		par->HorizDisplayEnd = (video_mode->xres + 3) * hmul / 8 - 1;
  wb_64 (regs, ECR_ADV_FUNC_CNTL, (TEXT ? 0x00 : 0x31));
#endif
	/* enhanced modes, linear addressing */
	s3_outw(trio, 0x4ae8, EC01_ENH_ENB | EC01_LAW_ENB, 0x4ae8);

	s3_crtc_outb(trio, 0x90, 0);

	/* H and V sync polarity */

	misc = s3_inb(trio, 0x3cc) & 0x2f;			/* colour, ram enable, clk sr12/s13 sel */
#ifdef CONFIG_NS1000
	if(video_mode->sync & FB_SYNC_HOR_HIGH_ACT)
		misc |= 0x40;					/* neg H sync polarity */
	if(video_mode->sync & FB_SYNC_VERT_HIGH_ACT)
		misc |= 0x80;					/* neg V sync polarity */
#else
	if(!(video_mode->sync & FB_SYNC_HOR_HIGH_ACT))
		misc |= 0x40;					/* neg H sync polarity */
	if(!(video_mode->sync & FB_SYNC_VERT_HIGH_ACT))
		misc |= 0x80;					/* neg V sync polarity */
#endif
	misc |= 0x0c;						/* clk from sr12/sr13 */
	s3_outb(trio, 0x3c2, misc);


	HDE = (video_mode->xres + 3) * hmul / 8 - 1;
	VDE = video_mode->yres - 1;
#if 0
	par->VertDisplayEnd = video_mode->yres - 1;
#endif


	s3_crtc_outb(trio, 0x45, 0);
	s3_crtc_outb(trio, 0x55, 0x10); /* 0=mswin 0x10=x11win */

	s3_seq_outb(trio, 0x04, (isTextMode || (video_mode->bits_per_pixel == 1)) ? 0x06 : 0x0e);

	s3_outb(trio, 0x3ce, 0x04);
	s3_outb(trio, 0x3cf, 0x00);

	s3_seq_outb(trio, 0x02, (video_mode->bits_per_pixel == 1) ? 0x01 : 0xFF);
	s3_seq_outb(trio, 0x03, 0x00);
	

	/* s3_compute_clock accepts arguments in Hz */
	/* pixclock is in ps ... convert to Hz */

	freq = (1000000000 / video_mode->pixclock) * 1000;
	debug_init("pixclock = %dps  freq = %ld\n", video_mode->pixclock, freq);
	mnr = s3_compute_clock(freq);
	s3_seq_outb(trio, 0x13, ((mnr & 0xFF00) >> 8));
	s3_seq_outb(trio, 0x12, (mnr & 0xFF));

	/* memory clock */
	mnr = s3_compute_clock(trio->memclk);
	s3_seq_outb(trio, 0x11, ((mnr & 0xFF00) >> 8));
	s3_seq_outb(trio, 0x10, (mnr & 0xFF));

	/* seq 0 - 4 */

	/* crt 0 - 24 */

	/* gfx 0 - 8 */

	/* atr 0 - 20 */

	/* trio order */


	/* Load display parameters into board */

	s3_crtc_outb(trio, 0x5d,
		((HT & 0x100) ? 0x01 : 0x00) |
		((HDE & 0x100) ? 0x02 : 0x00) |
		((HBS & 0x100) ? 0x04 : 0x00) |
		/* ((HBE & 0x40) ? 0x08 : 0x00) | */
		((HSS & 0x100) ? 0x10 : 0x00) |
		/* ((HSE & 0x20) ? 0x20 : 0x00) | */
		(((HT-5) & 0x100) ? 0x40 : 0x00)
		);

	s3_crtc_outb(trio, 0x5e,
		0x40 |
		((VT & 0x400) ? 0x01 : 0x00) |
		((VDE & 0x400) ? 0x02 : 0x00) |
		((VBS & 0x400) ? 0x04 : 0x00) |
		((VSS & 0x400) ? 0x10 : 0x00)
		);


	s3_crtc_outb(trio, 0x00, HT);
	s3_crtc_outb(trio, 0x3b, HT - 5);
	s3_crtc_outb(trio, 0x01, ((HDE >= HBS) ? (HBS - 1) : HDE));
	s3_crtc_outb(trio, 0x02, HBS);
	s3_crtc_outb(trio, 0x03, ((HBE & 0x1F) | 0x80));
	s3_crtc_outb(trio, 0x04, HSS);
	s3_crtc_outb(trio, 0x05, (HSE & 0x1F) | ((HBE & 0x20) ? 0x80 : 0x00));

	s3_crtc_outb(trio, 0x06, VT);
	s3_crtc_outb(trio, 0x07,
		0x10 |
		((VT & 0x100) ? 0x01 : 0x00) |
		((VDE & 0x100) ? 0x02 : 0x00) |
		((VSS & 0x100) ? 0x04 : 0x00) |
		((VBS & 0x100) ? 0x08 : 0x00) |
		((VT & 0x200) ? 0x20 : 0x00) |
		((VDE & 0x200) ? 0x40 : 0x00) |
		((VSS & 0x200) ? 0x80 : 0x00)
		);
	s3_crtc_outb(trio, 0x09,
		0x40 |
		(isDoubleScan ? 0x80 : 0x00) |
		((VBS & 0x200) ? 0x20 : 0x00) |
		(isTextMode ? ((fy - 1) & 0x1F) : 0x00)
		);
	
	s3_crtc_outb(trio, 0x17, 0xE3);


	s3_crtc_outb(trio, 0x0c, 0x00);
	s3_crtc_outb(trio, 0x0d, 0x00);
	s3_crtc_outb(trio, 0x10, VSS);
	s3_crtc_outb(trio, 0x11, (VSE & 0x0F) | 0x20);
	s3_crtc_outb(trio, 0x12, VDE);
	s3_crtc_outb(trio, 0x15, VBS);
	s3_crtc_outb(trio, 0x16, VBE);
	s3_crtc_outb(trio, 0x18, 0xFF);
	s3_crtc_outb(trio, 0x3c, HT / 2);
	s3_crtc_outb(trio, 0x42, (isInterlaced ? 0x20 : 0x00));

	s3_outb(trio, 0x3ce, 0x05);
	s3_outb(trio, 0x3cf, ((isTextMode || (video_mode->bits_per_pixel == 1)) ? 0x00 : 0x40));
	s3_outb(trio, 0x3ce, 0x06);
	s3_outb(trio, 0x3cf, (isTextMode ? 0x04 : 0x01));

	s3_seq_outb(trio, 0x04, ((isTextMode || (video_mode->bits_per_pixel == 1)) ? 0x06 : 0x02));
	
	/* DAC mask */
	s3_outw(trio, 0x3c6, 0xff, 0x3c6);

  /* Blank border */
  test = s3_crtc_inb(trio, 0x33);
  s3_crtc_outb(trio, 0x33, (test | 0x20));
	
	sr15 = 0x82;
	sr18 = 0;

  clock_mode = 0x00;
  cr50 = 0x00;
	cr53 = s3_crtc_inb(trio, 0x53) & 0xf9;
	
  test = s3_crtc_inb(trio, 0x67);
  test &= 0xD;
	
  switch (video_mode->bits_per_pixel) {
  case 8:
    if (freq > 80000000) {
      clock_mode = 0x10 | 0x02;
      sr15 |= 0x10;
      sr18 |= 0x80;
    }
    HDE = video_mode->xres / 8;
    cr50 |= 0x00;
    break;
		
  case 15:
    clock_mode = 0x30;
    HDE = video_mode->xres / 4;
    cr50 |= 0x10;
    cr53 |= 0x02;
    break;

  case 16:
    clock_mode = 0x50;
    HDE = video_mode->xres / 4;
    cr50 |= 0x10;
    cr53 |= 0x02;
    break;
		
  case 24:
    clock_mode = 0x70;
    HDE = video_mode->xres / 2;
    cr50 |= 0x30;
    cr53 |= 0x04;
    break;

  case 32:
    clock_mode = 0xD0;
    HDE = video_mode->xres / 2;
    cr50 |= 0x30;
    cr53 |= 0x04;
    break;
  }

  s3_crtc_outb(trio, 0x53, cr53);
  s3_crtc_outb(trio, 0x67, clock_mode | test);
  s3_seq_outb(trio, 0x15, sr15|0x03);
  s3_seq_outb(trio, 0x15, sr15|0x23);
  s3_seq_outb(trio, 0x15, sr15);
  s3_seq_outb(trio, 0x18, sr18);
  s3_crtc_outb(trio, 0x13, HDE);

	s3_crtc_outb(trio, 0x3a, (isTextMode ? 0x05 : 0x35));
	
  test = s3_crtc_inb(trio, 0x51);
  test &= ~0x30;
  test |= (HDE >> 4) & 0x30;
  s3_crtc_outb(trio, 0x51, test);


  /* Set up graphics engine */
  switch (video_mode->xres) {
  case 1024:
    cr50 |= 0x00;
    break;
  case 640:
    cr50 |= 0x40;
    break;
  case 800:
    cr50 |= 0x80;
    break;
  case 1280:
    cr50 |= 0xC0;
    break;
  case 1152:
    cr50 |= 0x01;
    break;
  case 1600:
    cr50 |= 0x81;
    break;
  default:	/* XXX */
    break;
  }

  s3_crtc_outb(trio, 0x50, cr50);
	
  udelay(100);
  s3_ac_outb(trio, 0x10, (isTextMode ? 0x08 : 0x41));
  udelay(100);
  s3_ac_outb(trio, 0x12, (video_mode->bits_per_pixel == 1) ? 0x01 : 0x0F);
  udelay(100);

  s3_ac_outb(trio, 0x20, 0);
  udelay(100);
	
  tfillm = (96 * (trio->memclk / 1000)) / 240000;
	
  switch (video_mode->bits_per_pixel) {
  case 32:
  case 24:
    temptym = (24 * (trio->memclk / 1000)) / (freq / 1000);
    break;
  case 15:
  case 16:
    temptym = (48 * (trio->memclk / 1000)) / (freq / 1000);
    break;
  case 4:
    temptym = (192 * (trio->memclk / 1000)) / (freq / 1000);
    break;
  default:
    temptym = (96 * (trio->memclk / 1000)) / (freq / 1000);
    break;
  }

  m = (temptym - tfillm - 9) / 2;
  if (m < 0)
    m = 0;
  m = (m & 0x1F) << 3;
  if (m < 0x18)
    m = 0x18;
  n = 0xFF;
	
	/* standard firmware setting and our driver differ */
	s3_crtc_outb(trio, 0x54, m);
	s3_crtc_outb(trio, 0x60, n);

	s3_display_enable(trio, 1);

	/* fixme: need the above palette code in some other place called from set_var */	

	debug_msg("s3_load_video_mode %dx%d-%dx%d-%dbpp %dps done\n",
		video_mode->xres, video_mode->yres, video_mode->xres_virtual, video_mode->yres_virtual,
		video_mode->bits_per_pixel, video_mode->pixclock);

	return(0);
}


static int s3_set_var(triofb_t *trio, int con, struct fb_var_screeninfo *new_var)
{
	struct display *display;
	struct fb_var_screeninfo *cur_var;

	debug_flow("s3_set_var con=%d\n", con);
	display = (con < 0) ? &trio->disp : &fb_display[con];
	cur_var = &display->var;

	if(cur_var->xres != new_var->xres || cur_var->yres != new_var->yres
	|| cur_var->pixclock != new_var->pixclock
	|| cur_var->left_margin != new_var->left_margin || cur_var->right_margin != new_var->right_margin
	|| cur_var->upper_margin != new_var->upper_margin || cur_var->lower_margin != new_var->lower_margin
	|| cur_var->xres_virtual != new_var->xres_virtual || cur_var->yres_virtual != new_var->yres_virtual
	|| cur_var->bits_per_pixel != new_var->bits_per_pixel || cur_var->accel_flags != new_var->accel_flags) {

		debug_flow("s3_set_var updating display->var\n");
		display->var = *new_var;

		display->screen_base = trio->vram_virt;
		display->dispsw = &fbcon_dummy;
		display->type = FB_TYPE_PACKED_PIXELS;
		display->type_aux = 0;
		display->ypanstep = 1;
		display->ywrapstep = 0;
		display->can_soft_blank = 1;
		display->inverse = 0;
		if(trio->accel) {
			cur_var->accel_flags |= FB_ACCELF_TEXT;
			display->scrollmode = 0;
		} else {
			cur_var->accel_flags &= ~FB_ACCELF_TEXT;
			display->scrollmode = SCROLL_YREDRAW;
		}
		display->visual = FB_VISUAL_DIRECTCOLOR;
		display->visual = FB_VISUAL_TRUECOLOR;

		if(cur_var->bits_per_pixel != 15) {
			cur_var->bits_per_pixel = (cur_var->bits_per_pixel + 7) & ~7;
		}
		cur_var->red.offset = cur_var->green.offset = cur_var->blue.offset = 0;
		cur_var->transp.offset = 0;
		cur_var->red.length = cur_var->green.length = cur_var->blue.length = 8;
		cur_var->transp.length = 0;

#ifdef FBCON_HAS_CFB8
		if(cur_var->bits_per_pixel == 8) {
			if(display->var.accel_flags & FB_ACCELF_TEXT)
				display->dispsw = &s3triofb_cfb8;
			else
				display->dispsw = &fbcon_cfb8;
			display->line_length = cur_var->xres_virtual;
			display->visual = FB_VISUAL_PSEUDOCOLOR;
			cur_var->red.offset = 0;
			cur_var->green.offset = 0;
			cur_var->blue.offset = 0;
			cur_var->red.length = cur_var->green.length = cur_var->blue.length = 6;
		}
#endif

#ifdef FBCON_HAS_CFB16
		if(cur_var->bits_per_pixel == 15) {
			if(display->var.accel_flags & FB_ACCELF_TEXT)
				display->dispsw = &s3triofb_cfb16;
			else
				display->dispsw = &fbcon_cfb16;
			display->dispsw_data = trio->con_cmap.cfb16;
			display->line_length = cur_var->xres_virtual * 2;
			cur_var->red.offset = 10;
			cur_var->green.offset = 5;
			cur_var->blue.offset = 0;
			cur_var->red.length = cur_var->green.length = cur_var->blue.length = 5;
		}
		if(cur_var->bits_per_pixel == 16) {
			if(display->var.accel_flags & FB_ACCELF_TEXT)
				display->dispsw = &s3triofb_cfb16;
			else
				display->dispsw = &fbcon_cfb16;
			display->dispsw_data = trio->con_cmap.cfb16;
			display->line_length = cur_var->xres_virtual * 2;
			cur_var->red.offset = 11;
			cur_var->green.offset = 5;
			cur_var->blue.offset = 0;
			cur_var->red.length = cur_var->blue.length = 5;
			cur_var->green.length = 6;
		}
#endif

#ifdef FBCON_HAS_CFB24
		if(cur_var->bits_per_pixel == 24) {
			display->dispsw = &fbcon_cfb24;
			display->dispsw_data = trio->con_cmap.cfb24;
			display->line_length = cur_var->xres_virtual * 3;
			cur_var->red.offset = 16;
			cur_var->green.offset = 8;
			cur_var->blue.offset = 0;
			cur_var->red.length = cur_var->green.length = cur_var->blue.length = 8;
		}
#endif

#ifdef FBCON_HAS_CFB32
		if(cur_var->bits_per_pixel == 32) {
#if 0
			if(display->var.accel_flags & FB_ACCELF_TEXT)
				display->dispsw = &s3triofb_cfb32;
			else
#endif
				display->dispsw = &fbcon_cfb32;
			display->dispsw_data = trio->con_cmap.cfb32;
			display->line_length = cur_var->xres_virtual * 4;
			cur_var->transp.offset = 24;
			cur_var->red.offset = 16;
			cur_var->green.offset = 8;
			cur_var->blue.offset = 0;
			cur_var->red.length = cur_var->green.length =
			cur_var->blue.length = cur_var->transp.length = 8;
		}
#endif


#if 0
	if (cur_var->xres * nom / den * cur_var->yres > s3info->s3hw.videoRam) {
		printk ("s3: mode %dx%dx%d rejected...resolution too high to fit into video memory!\n",
			new_var->xres, new_var->yres, new_var->bits_per_pixel);
		DPRINTK ("EXIT - EINVAL error\n");
		return -EINVAL;
	}

	/* use highest possible virtual resolution */
	if (cur_var->xres_virtual == -1 &&
	    cur_var->yres_virtual == -1) {
		printk ("s3: using maximum available virtual resolution\n");
		for (i = 0; modes[i].xres != -1; i++) {
			if (modes[i].xres * nom / den * modes[i].yres < s3info->s3hw.videoRam / 2)
				break;
		}
		if (modes[i].xres == -1) {
			printk ("s3: could not find a virtual resolution that fits into video memory!!\n");
			DPRINTK ("EXIT - EINVAL error\n");
			return -EINVAL;
		}
		cur_var->xres_virtual = modes[i].xres;
		cur_var->yres_virtual = modes[i].yres;

		printk ("s3: virtual resolution set to maximum of %dx%d\n",
			cur_var->xres_virtual, cur_var->yres_virtual);
	} else if (cur_var->xres_virtual == -1) {
		/* FIXME: maximize X virtual resolution only */
	} else if (cur_var->yres_virtual == -1) {
		/* FIXME: maximize Y virtual resolution only */
	}
#endif

	if (cur_var->xoffset < 0)
		cur_var->xoffset = 0;
	if (cur_var->yoffset < 0)
		cur_var->yoffset = 0;

	/* truncate xoffset and yoffset to maximum if too high */
	if (cur_var->xoffset > cur_var->xres_virtual - cur_var->xres)
		cur_var->xoffset = cur_var->xres_virtual - cur_var->xres - 1;

	if (cur_var->yoffset > cur_var->yres_virtual - cur_var->yres)
		cur_var->yoffset = cur_var->yres_virtual - cur_var->yres - 1;

	cur_var->red.msb_right =
	    cur_var->green.msb_right =
	    cur_var->blue.msb_right =
	    cur_var->transp.msb_right = 0;

#if 0
#warning FIXME: verify that the above code sets dsp->* fields correctly
#endif


		if(con == trio->currcon) {
			debug_flow("s3_set_var updating hardware\n");
#if TRIO_DUMP_REGS > 1
			trio_dump_regs(trio, "before load_video_mode");
#endif
			s3_load_video_mode(trio, &display->var);
#if TRIO_DUMP_REGS > 1
			trio_dump_regs(trio, "after load_video_mode");
#endif
			do_install_cmap(con, &trio->fb_info);
		}


		if(trio->fb_info.changevar)
			trio->fb_info.changevar(con);
	}

	return(0);
}


    /*
     *  Get the Fixed Part of the Display
     */

static int s3triofb_op_get_fix(struct fb_fix_screeninfo *fix, int con,
			  struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	struct display *disp;
        
	debug_flow("s3triofb_op_get_fix con=%d\n", con);

	disp = (con < 0) ? trio->fb_info.disp : &fb_display[con];

	memset(fix, 0, sizeof (struct fb_fix_screeninfo));
	strcpy(fix->id, trio->name);
        
	fix->smem_start = trio->vram_phys;
	fix->smem_len = trio->vram_size;

	fix->type = disp->type;
	fix->type_aux = disp->type_aux;
	fix->visual = disp->visual;

	fix->xpanstep = 1;
	fix->ypanstep = 1;
	fix->ywrapstep = 0;
        
	fix->line_length = disp->line_length;
 
	fix->mmio_start = trio->mmio_phys;
	fix->mmio_len = trio->mmio_size;
	fix->accel = trio->accel ? FB_ACCEL_S3_TRIO64 : FB_ACCEL_NONE;
	return 0;
}


    /*
     *  Get the User Defined Part of the Display
     */

static int s3triofb_op_get_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;

	debug_flow("s3triofb_op_get_var con=%d\n", con);
debug_flow("cr82e8=0x%04x cr86e8=0x%04x\n", s3_inw(trio, 0x82e8), s3_inw(trio, 0x86e8));
	*var = (con < 0) ? trio->disp.var : fb_display[con].var;

	return 0;
}


    /*
     *  Set the User Defined Part of the Display
     */

static int s3triofb_op_set_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	int activate, bytespp;

	activate = var->activate & FB_ACTIVATE_MASK;

	debug_msg("s3triofb_op_set_var con=%d %dx%d-%dx%d-%dbpp mode=%d %d %d %d %d %d %d %s act=%d\n",
		con, var->xres, var->yres, var->xres_virtual, var->yres_virtual, var->bits_per_pixel,
		var->pixclock, var->left_margin, var->right_margin, var->upper_margin, var->lower_margin,
		var->hsync_len, var->vsync_len,
		var->accel_flags & FB_ACCELF_TEXT ? "AT" : "--",
		activate);

	bytespp = (var->bits_per_pixel + 7) >> 3;
	if(var->xres_virtual == -1)
		var->xres_virtual = var->xres;
	if(var->yres_virtual == -1)
		var->yres_virtual = var->yres;

	if(!(var->bits_per_pixel == 8 || var->bits_per_pixel == 15 || var->bits_per_pixel == 16
	|| /* var->bits_per_pixel == 24 || */ var->bits_per_pixel == 32)) {
		return(-EINVAL);
	}

	if(var->xres * var->yres * bytespp > trio->vram_size) {
		return(-ENOMEM);
	}

#if 0
	if(activate == FB_ACTIVATE_NXTOPEN)
		fixme?
#endif

	if(activate == FB_ACTIVATE_NOW) {

		s3_set_var(trio, con, var);
	}

	return(0);
}


    /*
     *  Pan or Wrap the Display
     *
     *  This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
     */

static int s3triofb_op_pan_display(struct fb_var_screeninfo *var, int con,
			      struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	unsigned int base;

	debug_flow("s3triofb_op_pan_display con=%d\n", con);
	if(var->xoffset > (var->xres_virtual - var->xres))
		return -EINVAL;
	if(var->yoffset > (var->yres_virtual - var->yres))
		return -EINVAL;

#if 0
	fb_var.xoffset = var->xoffset;
	fb_var.yoffset = var->yoffset;
	base = var->yoffset * fb_fix.line_length + var->xoffset;
#endif

        base = ((var->yoffset * var->xres + var->xoffset) * var->bits_per_pixel) >> 3;

#if 0
	outw(IO_OUT16VAL((base >> 8) & 0xff, 0x0c),0x03D4);
	outw(IO_OUT16VAL(base  & 0xff, 0x0d),0x03D4);
	outw(IO_OUT16VAL((base >> 16) & 0xf, 0x69),0x03D4);
#endif
	s3_crtc_outb(trio, 0x0c, base >> 8);
	s3_crtc_outb(trio, 0x0d, base);
	s3_crtc_outb(trio, 0x69, (base >> 16) & 0xf);

	return 0;
}


    /*
     *  Get the Colormap
     */

static int s3triofb_op_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	debug_flow("s3triofb_op_get_cmap con=%d\n", con);
	if (con == trio->currcon) /* current console? */
		return fb_get_cmap(cmap, kspc, s3trio_getcolreg, info);
	else if (fb_display[con].cmap.len) /* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(1 << fb_display[con].var.bits_per_pixel),
				cmap, kspc ? 0 : 2);
	return 0;
}

    /*
     *  Set the Colormap
     */

static int s3triofb_op_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	int err;

	debug_flow("s3triofb_op_set_cmap con=%d\n", con);

	/* fixme: this is static for us, no? */
	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		debug_flow("s3triofb_op_set_cmap con=%d allocating colormap\n", con);
		if((err = fb_alloc_cmap(&fb_display[con].cmap, 1<<fb_display[con].var.bits_per_pixel, 0)))
			return err;
	}

	if(con == trio->currcon) {
		return(fb_set_cmap(cmap, kspc, s3trio_setcolreg, info));
	}

	fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}





static int __init s3triofb_pci_init(struct pci_dev *pdev)
{
	triofb_t *trio;
	int i;


	trio = &xox;	/* fixme: kmalloc() */

	memset(trio, 0, sizeof(triofb_t));
	trio->pdev = pdev;
	trio->accel = 1;

#if TRIO_DUMP_PCICFG
	trio_dump_pcicfg(pdev);
#endif
	if(pci_enable_device(pdev)) {
		printk("s3triofb: cannot enable device\n");
		return -ENODEV;
	}

#ifdef CONFIG_ALL_PPC
	printk("CONFIG_ALL_PPC\n");
#endif
#ifdef CONFIG_PPC
	printk("CONFIG_PPC\n");
#endif

	strcpy(trio->name, "s3trio");

	/* PCI configuration fix-up */

	trio->bar0_phys = pci_resource_start(pdev, 0);
	trio->bar0_size = pci_resource_len(pdev, 0);

	if(!request_mem_region(trio->bar0_phys, trio->bar0_size, "s3triofb")) {
		printk("s3triofb_pci_init: request_mem_region failed.\n");
		return(-ENODEV);
	}
	trio->bar0_virt = ioremap(trio->bar0_phys, trio->bar0_size);
	debug_msg("s3triofb: bar0_virt = 0x%x (%dk)\n", (uint)trio->bar0_virt, trio->bar0_size/1024);

	trio->mmio_phys = trio->bar0_phys + 0x1000000;
	trio->mmio_size = 0x1000000;
	trio->mmio_virt = trio->bar0_virt + 0x1000000;
	trio->mmio_regs = trio->bar0_virt + 0x1008000;
	debug_msg("s3triofb: mmio_phys=0x%x mmio_virt=0x%x mmio_regs=0x%x\n", (uint)trio->mmio_phys, (uint)trio->mmio_virt, (uint)trio->mmio_regs);

#define VRAMOFFSET 32 * 1024 * 1024
#define _VRAMOFFSET 0

	trio->vram_phys = trio->bar0_phys + VRAMOFFSET;
	trio->vram_virt = trio->bar0_virt + VRAMOFFSET;
	trio->vram_size = 2 * 1024 * 1024 - 0x400;
	debug_msg("s3triofb: vram_phys=0x%x vram_virt=0x%x\n", (uint)trio->vram_phys, (uint)trio->vram_virt);


	/* Enable IO and memory-mapped card access */
	pci_write_config_word(pdev, PCI_COMMAND, PCI_COMMAND_IO | PCI_COMMAND_MEMORY);

	pci_set_drvdata(pdev, trio);

	/* setup fb_info member */

	strcpy(trio->fb_info.modename, "s3triofb_mode");
	trio->fb_info.node = -1;
	trio->fb_info.flags = FBINFO_FLAG_DEFAULT;
	trio->fb_info.fbops = &s3trio_ops;
	trio->fb_info.display_fg = NULL;
#if 0
	trio->fb_info.fbvar_num = 1;
	trio->fb_info.fbvar = &fb_var;
#endif
	trio->fb_info.disp = &trio->disp;	/* fixme: is this changed later on? */
	trio->fb_info.fontname[0] = '\0';

	trio->fb_info.changevar = NULL;
	trio->fb_info.switch_con = &s3triofbcon_switch;
	trio->fb_info.updatevar = &s3triofbcon_updatevar;
	trio->fb_info.blank = &s3triofbcon_blank;
#if 0
	trio->fb_info.setcmap = &s3triofbcon_setcmap;
#endif


	trio->currcon = -1;
	trio->currcon_display = &trio->disp;

	/* fixme: add save state */

	/* setup default color table */
	for(i = 0; i < 16; i++) {
		int j = color_table[i];
		trio->palette[i].red=default_red[j];
		trio->palette[i].green=default_grn[j];
		trio->palette[i].blue=default_blu[j];
	}

	s3_unlock(trio);
#if TRIO_DUMP_REGS
	trio_dump_regs(trio, "early init");
#endif

	s3_setup(trio);

	/* determine vram size */
	trio->vram_size = 0x200000;
	trio->memclk = 50000000;
	trio->memclk = 60000000;

	s3_set_var(trio, trio->currcon, &s3triofb_default_var);

#if 0
#ifdef CONFIG_FB_COMPAT_XPMAC
	debug_flow("s3triofb: CONFIG_FB_COMPAT_XPMAC is defined\n");
	if (!console_fb_info) {
		display_info.height = fb_var.yres;
		display_info.width = fb_var.xres;
		display_info.depth = 8;
		display_info.pitch = fb_fix.line_length;
		display_info.mode = 0;
		strncpy(display_info.name, fb_info.modename, sizeof(display_info.name));
		display_info.fb_address = (unsigned long)fb_fix.smem_start;
		display_info.disp_reg_address = mmio_regs;
		display_info.cmap_adr_address = mmio_regs + 0x3c8;
		display_info.cmap_data_address = mmio_regs + 0x3c9;
		console_fb_info = &fb_info;
	}
#endif /* CONFIG_FB_COMPAT_XPMAC) */
#endif

	debug_flow("s3triofb: about to register fbdev\n");

	if(register_framebuffer((struct fb_info *)trio) < 0) {
printk("trio_pci_init: register_framebuffer failed\n");
		return(-ENODEV);
	}

#if TRIO_DUMP_REGS
	trio_dump_regs(trio, "init complete");
#endif

	printk("fb%d: %s frame buffer device\n",
		GET_FB_IDX(trio->fb_info.node), trio->fb_info.modename);
	return(0);
}



static int s3triofbcon_switch(int con, struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	struct display *dsp;

	debug_flow("s3triofb_con_switch con=%d\n", con);

	dsp = (con < 0) ? &trio->disp : &fb_display[con];

	if(trio->currcon >= 0) {
		/* Do we have to save the colormap? */
		if (fb_display[trio->currcon].cmap.len)
			fb_get_cmap(&fb_display[trio->currcon].cmap, 1, s3trio_getcolreg, info);
	}

	trio->currcon = con;
	trio->currcon_display = dsp;

	/* Install new colormap */
	do_install_cmap(con, info);
	return 0;
}

    /*
     *  Update the `var' structure (called by fbcon.c)
     */

static int s3triofbcon_updatevar(int con, struct fb_info *info)
{
	debug_flow("s3triofb_con_updatevar con=%d\n", con);
	/* Nothing */
	return 0;
}

    /*
     *  Blank the display.
     */

static void s3triofbcon_blank(int blank, struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	uint8_t sr1, srd;
	debug_flow("s3triofb_con_blank %d\n", blank);

#if 0
	s3_mmio_put8(trio, 0x3c4, 0x1, trio->mmio_regs + 0x03c4);
	x = s3_mmio_get8(trio, 0x3c5);
	s3_mmio_put8(trio, 0x3c5, (x & (~0x20)) | (blank << 5), trio->mmio_regs + 0x03c5);
#endif
	sr1 = s3_seq_inb(trio, 0x01) & ~0x20;
	srd = s3_seq_inb(trio, 0x0d) & 0x0f;
	if(blank) {
		sr1 |= 0x20;
		switch(blank - 1) {
		default:
		case VESA_NO_BLANKING:
			break;
		case VESA_VSYNC_SUSPEND:
			srd |= 0x40;
			break;
		case VESA_HSYNC_SUSPEND:
			srd |= 0x10;
			break;
		case VESA_POWERDOWN:
			srd |= 0x50;
			break;
		}
	}
	s3_seq_outb(trio, 0x0d, srd);
	s3_seq_outb(trio, 0x01, sr1);
}

    /*
     *  Set the colormap
     */

#if 0
static int s3triofbcon_setcmap(struct fb_cmap *cmap, int con)
{
    return(s3trio_set_cmap(cmap, 1, con, &fb_info));
}
#endif


    /*
     *  Read a single color register and split it into
     *  colors/transparent. Return != 0 for invalid regno.
     */

static int s3trio_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	if (regno > 255)
		return 1;
	*red = (trio->palette[regno].red << 8) | trio->palette[regno].red;
	*green = (trio->palette[regno].green << 8) | trio->palette[regno].green;
	*blue = (trio->palette[regno].blue << 8) | trio->palette[regno].blue;
	*transp = 0;
	return 0;
}


    /*
     *  Set a single color register. Return != 0 for invalid regno.
     */

static int s3trio_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
		u_int transp, struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
	struct display *dsp = trio->currcon_display;

	if (regno > 255)
		return 1;

	red >>= 8;
	green >>= 8;
	blue >>= 8;
	transp >>= 8;

	if(regno < 4) {
		debug_flow("s3trio_setcolreg %3d rgba=%02x.%02x.%02x.%02x\n", regno, red, green, blue, transp);
	}

	trio->palette[regno].red = red;
	trio->palette[regno].green = green;
	trio->palette[regno].blue = blue;

	switch(dsp->var.bits_per_pixel) {
#ifdef FBCON_HAS_CFB8
	case 8:
		s3_mmio_put8(trio, 0x3c8, regno, trio->mmio_regs + 0x3c8);
		s3_mmio_put8(trio, 0x3c9, (red & 0xff) >> 2, trio->mmio_regs + 0x3c9);
		s3_mmio_put8(trio, 0x3c9, (green & 0xff) >> 2, trio->mmio_regs + 0x3c9);
		s3_mmio_put8(trio, 0x3c9, (blue & 0xff) >> 2, trio->mmio_regs + 0x3c9);
	break;
#endif
#ifdef FBCON_HAS_CFB16	
	case 15:
		if(regno < 16) {
			trio->con_cmap.cfb16[regno] =
				((red >> 3) << 10) | ((green >> 3) << 5) | ((blue >> 3));
		}
		break;

	case 16:
		if(regno < 16) {
			trio->con_cmap.cfb16[regno] =
				((red >> 3) << 11) | ((green >> 2) << 5) | ((blue >> 3));
		}
		break;
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
		if(regno < 16)
		trio->con_cmap.cfb24[regno] =
			(red   << dsp->var.red.offset)   |
			(green << dsp->var.green.offset) |
			(blue  << dsp->var.blue.offset);
		break;
#endif
#ifdef FBCON_HAS_CFB32
	case 32:
		if(regno < 16)
		trio->con_cmap.cfb32[regno] =
			(red   << dsp->var.red.offset)   |
			(green << dsp->var.green.offset) |
			(blue  << dsp->var.blue.offset)  |
			(transp << dsp->var.transp.offset);	/* 8:8:8:8 */
		break;
#endif
	default:
		/* should not occur */
		break;
	}

	return 0;
}


static void do_install_cmap(int con, struct fb_info *info)
{
	triofb_t *trio = (triofb_t*)info;
/* fixme: can con be < 0 ? */
	debug_flow("s3_do_install_cmap con=%d\n", con);
	if(con != trio->currcon || con < 0)
		return;
	if (fb_display[con].cmap.len) {
		fb_set_cmap(&fb_display[con].cmap, 1, s3trio_setcolreg, &trio->fb_info);
	} else {
		fb_set_cmap(fb_default_cmap(fb_display[con].var.bits_per_pixel), 1,
				s3trio_setcolreg, &trio->fb_info);
	}
}

static void Trio_WaitQueue(triofb_t *trio, u_short fifo)
{
	u_short status;

        do {
		status = s3_mmio_get16(trio, 0x1AE8);
	} while(!(status & fifo));
}

static void Trio_WaitBlit(triofb_t *trio)
{
	u_short status;

        do {
		status = s3_mmio_get16(trio, 0x1AE8);
	} while(status & 0x200);
}

static void Trio_BitBLT(triofb_t *trio, u_short curx, u_short cury,
	u_short destx, u_short desty, u_short width, u_short height, u_short mode)
{
	u_short blitcmd = 0xc011;

	/* Set drawing direction */
        /* -Y, X maj, -X (default) */

	if (curx > destx)
		blitcmd |= 0x0020;  /* Drawing direction +X */
	else {
		curx  += (width - 1);
		destx += (width - 1);
	}

	if (cury > desty)
		blitcmd |= 0x0080;  /* Drawing direction +Y */
	else {
		cury  += (height - 1);
		desty += (height - 1);
	}

	Trio_WaitQueue(trio, 0x0400);

	s3_outw(trio, 0xbee8, 0xa000,  0xBEE8);
	s3_outw(trio, 0xbae8, 0x60 | mode,  0xBAE8);

	s3_outw(trio, 0x86e8, curx,  0x86E8);
	s3_outw(trio, 0x82e8, cury,  0x82E8);

	s3_outw(trio, 0x8ee8, destx,  0x8EE8);
	s3_outw(trio, 0x8ae8, desty,  0x8AE8);

	s3_outw(trio, 0xbee8, height - 1,  0xBEE8);
	s3_outw(trio, 0x96e8, width - 1,  0x96E8);

	s3_outw(trio, 0x9ae8, blitcmd, 0x9AE8);

}

static void Trio_RectFill(triofb_t *trio, u_short x, u_short y,
	u_short width, u_short height, u_short mode, u_short color)
{
	u_short blitcmd = 0x40b1;

	Trio_WaitQueue(trio, 0x0400);

	s3_outw(trio, 0xbee8, 0xa000,  0xBEE8);
	s3_outw(trio, 0xbae8, (0x20 | mode),  0xBAE8);
	s3_outw(trio, 0xbee8, 0xe000,  0xBEE8);
	s3_outw(trio, 0xa6e8, color,  0xA6E8);
	s3_outw(trio, 0x86e8, x,  0x86E8);
	s3_outw(trio, 0x82e8, y,  0x82E8);
	s3_outw(trio, 0xbee8, (height - 1), 0xBEE8);
	s3_outw(trio, 0x96e8, (width - 1), 0x96E8);
	s3_outw(trio, 0x9ae8, blitcmd,  0x9AE8);

}


static void Trio_MoveCursor(triofb_t *trio, u_short x, u_short y)
{
#if 0
	mem_out8(0x39, xox.mmio_regs + 0x3d4);
	mem_out8(0xa0, xox.mmio_regs + 0x3d5);

	mem_out8(0x46, xox.mmio_regs + 0x3d4);
	mem_out8((x & 0x0700) >> 8, xox.mmio_regs + 0x3d5);
	mem_out8(0x47, xox.mmio_regs + 0x3d4);
	mem_out8(x & 0x00ff, xox.mmio_regs + 0x3d5);

	mem_out8(0x48, xox.mmio_regs + 0x3d4);
	mem_out8((y & 0x0700) >> 8, xox.mmio_regs + 0x3d5);
	mem_out8(0x49, xox.mmio_regs + 0x3d4);
	mem_out8(y & 0x00ff, xox.mmio_regs + 0x3d5);
#endif
	s3_crtc_write(trio, 0x39, 0xa0);
	s3_crtc_write(trio, 0x46, (x & 0x0700) >> 8);
	s3_crtc_write(trio, 0x47, x & 0x00ff);
	s3_crtc_write(trio, 0x48, (y & 0x0700) >> 8);
	s3_crtc_write(trio, 0x49, y & 0x00ff);
}


    /*
     *  Text console acceleration
     */

#ifdef FBCON_HAS_CFB8
static void s3triofb_cfb8_bmove(struct display *p, int sy, int sx, int dy,
		int dx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	sx *= 8; dx *= 8; width *= 8;
	Trio_BitBLT(trio, (u_short)sx, (u_short)(sy*fontheight(p)), (u_short)dx,
		(u_short)(dy*fontheight(p)), (u_short)width,
		(u_short)(height*fontheight(p)), (u_short)S3_NEW);
}

static void s3triofb_cfb8_clear(struct vc_data *conp, struct display *p, int sy,
		int sx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	uint8_t bg;

	sx *= 8; width *= 8;
	bg = attr_bgcol_ec(p,conp);
	Trio_RectFill(trio, (u_short)sx, (u_short)(sy*fontheight(p)),
		(u_short)width, (u_short)(height*fontheight(p)),
		(u_short)S3_NEW, (u_short)bg);
}

static void s3triofb_cfb8_putc(struct vc_data *conp, struct display *p, int c,
		int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb8_putc(conp, p, c, yy, xx);
}

static void s3triofb_cfb8_putcs(struct vc_data *conp, struct display *p,
		const unsigned short *s, int count, int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb8_putcs(conp, p, s, count, yy, xx);
}

static void s3triofb_cfb8_revc(struct display *p, int xx, int yy)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb8_revc(p, xx, yy);
}

static struct display_switch s3triofb_cfb8 = {
	setup:		fbcon_cfb8_setup,
	bmove:		s3triofb_cfb8_bmove,
	clear:		s3triofb_cfb8_clear,
	putc:		s3triofb_cfb8_putc,
	putcs:		s3triofb_cfb8_putcs,
	revc:		s3triofb_cfb8_revc,
	clear_margins:	fbcon_cfb8_clear_margins,
	fontwidthmask:	FONTWIDTH(8)
};
#endif


#ifdef FBCON_HAS_CFB16
static void s3triofb_cfb16_bmove(struct display *p, int sy, int sx, int dy,
		int dx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	sx *= 8; dx *= 8; width *= 8;
	Trio_BitBLT(trio, (u_short)sx, (u_short)(sy*fontheight(p)), (u_short)dx,
		(u_short)(dy*fontheight(p)), (u_short)width,
		(u_short)(height*fontheight(p)), (u_short)S3_NEW);
}

static void s3triofb_cfb16_clear(struct vc_data *conp, struct display *p, int sy,
		int sx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	uint8_t bg;

	sx *= 8; width *= 8;
	bg = attr_bgcol_ec(p,conp);
	Trio_RectFill(trio, (u_short)sx, (u_short)(sy*fontheight(p)),
		(u_short)width, (u_short)(height*fontheight(p)),
		(u_short)S3_NEW, (u_short)bg);
}

static void s3triofb_cfb16_putc(struct vc_data *conp, struct display *p, int c,
		int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb16_putc(conp, p, c, yy, xx);
}

static void s3triofb_cfb16_putcs(struct vc_data *conp, struct display *p,
		const unsigned short *s, int count, int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb16_putcs(conp, p, s, count, yy, xx);
}

static void s3triofb_cfb16_revc(struct display *p, int xx, int yy)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb16_revc(p, xx, yy);
}

static struct display_switch s3triofb_cfb16 = {
	setup:		fbcon_cfb16_setup,
	bmove:		s3triofb_cfb16_bmove,
	clear:		s3triofb_cfb16_clear,
	putc:		s3triofb_cfb16_putc,
	putcs:		s3triofb_cfb16_putcs,
	revc:		s3triofb_cfb16_revc,
	clear_margins:	fbcon_cfb16_clear_margins,
	fontwidthmask:	FONTWIDTH(8)
};
#endif


#ifdef FBCON_HAS_CFB32
static void s3triofb_cfb32_bmove(struct display *p, int sy, int sx, int dy,
		int dx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	sx *= 8; dx *= 8; width *= 8;
	Trio_BitBLT(trio, (u_short)sx, (u_short)(sy*fontheight(p)), (u_short)dx,
		(u_short)(dy*fontheight(p)), (u_short)width,
		(u_short)(height*fontheight(p)), (u_short)S3_NEW);
}

static void s3triofb_cfb32_clear(struct vc_data *conp, struct display *p, int sy,
		int sx, int height, int width)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	uint8_t bg;

	sx *= 8; width *= 8;
	bg = attr_bgcol_ec(p,conp);
	Trio_RectFill(trio, (u_short)sx, (u_short)(sy*fontheight(p)),
		(u_short)width, (u_short)(height*fontheight(p)),
		(u_short)S3_NEW, (u_short)bg);
}

static void s3triofb_cfb32_putc(struct vc_data *conp, struct display *p, int c,
		int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb32_putc(conp, p, c, yy, xx);
}

static void s3triofb_cfb32_putcs(struct vc_data *conp, struct display *p,
		const unsigned short *s, int count, int yy, int xx)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb32_putcs(conp, p, s, count, yy, xx);
}

static void s3triofb_cfb32_revc(struct display *p, int xx, int yy)
{
	triofb_t *trio = (triofb_t*)p->fb_info;
	Trio_WaitBlit(trio);
	fbcon_cfb32_revc(p, xx, yy);
}

static struct display_switch s3triofb_cfb32 = {
	setup:		fbcon_cfb32_setup,
	bmove:		s3triofb_cfb32_bmove,
	clear:		s3triofb_cfb32_clear,
	putc:		s3triofb_cfb32_putc,
	putcs:		s3triofb_cfb32_putcs,
	revc:		s3triofb_cfb32_revc,
	clear_margins:	fbcon_cfb32_clear_margins,
	fontwidthmask:	FONTWIDTH(8)
};
#endif




int __init s3triofb_init (void)
{
	debug_flow("s3triofb_init\n");
	return pci_module_init (&s3triofb_driver);
}


void __exit s3triofb_exit (void)
{
	debug_flow("s3triofb_exit\n");
	pci_unregister_driver (&s3triofb_driver);
}


int __init s3triofb_setup (char *options)
{
	debug_flow("s3triofb_setup: options='%s'\n", options ? options : "null");
#if 0
        char *this_opt;

        if (!options || !*options)
                return 0;
 
	while ((this_opt = strsep (&options, ",")) != NULL) {
		if (!*this_opt)
			continue;
                if (!strncmp (this_opt, "font:", 5)) {
                        char *p;
                        int i;
        
                        p = this_opt + 5;
                        for (i=0; i<sizeof (fontname) - 1; i++)
                                if (!*p || *p == ' ' || *p == ',')
                                        break;
                        memcpy(fontname, this_opt + 5, i);
                } else if (!strncmp(this_opt, "noaccel", 7)) {
			noaccel = 1;
		} else if (!strncmp(this_opt, "dfp", 3)) {
			force_dfp = 1;
		} else if (!strncmp(this_opt, "panel_yres:", 11)) {
			panel_yres = simple_strtoul((this_opt+11), NULL, 0);
                } else
			mode_option = this_opt;
        }
#endif

	return 0;
}

static int s3triofb_pci_register (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	s3triofb_pci_init(pdev);

#if 0
	struct s3triofb_info *rinfo;
	u32 tmp;
	int i, j;

	RTRACE("s3triofb_pci_register BEGIN\n");

	rinfo = kmalloc (sizeof (struct s3triofb_info), GFP_KERNEL);
	if (!rinfo) {
		printk ("s3triofb: could not allocate memory\n");
		return -ENODEV;
	}

	memset (rinfo, 0, sizeof (struct s3triofb_info));

	rinfo->pdev = pdev;

	/* enable device */
	{
		int err;

		if ((err = pci_enable_device(pdev))) {
			printk("s3triofb: cannot enable device\n");
			kfree (rinfo);
			return -ENODEV;
		}
	}

	/* set base addrs */
	rinfo->fb_base_phys = pci_resource_start (pdev, 0);
	rinfo->mmio_base_phys = pci_resource_start (pdev, 2);

	/* request the mem regions */
	if (!request_mem_region (rinfo->fb_base_phys,
				 pci_resource_len(pdev, 0), "s3triofb")) {
		printk ("s3triofb: cannot reserve FB region\n");
		kfree (rinfo);
		return -ENODEV;
	}

	if (!request_mem_region (rinfo->mmio_base_phys,
				 pci_resource_len(pdev, 2), "s3triofb")) {
		printk ("s3triofb: cannot reserve MMIO region\n");
		release_mem_region (rinfo->fb_base_phys,
				    pci_resource_len(pdev, 0));
		kfree (rinfo);
		return -ENODEV;
	}

	/* map the regions */
	rinfo->mmio_base = (u32) ioremap (rinfo->mmio_base_phys,
				    		    RADEON_REGSIZE);
	if (!rinfo->mmio_base) {
		printk ("s3triofb: cannot map MMIO\n");
		release_mem_region (rinfo->mmio_base_phys,
				    pci_resource_len(pdev, 2));
		release_mem_region (rinfo->fb_base_phys,
				    pci_resource_len(pdev, 0));
		kfree (rinfo);
		return -ENODEV;
	}

	rinfo->chipset = pdev->device;

	/* chipset */
	switch (pdev->device) {
		case PCI_DEVICE_ID_RADEON_QD:
			strcpy(rinfo->name, "Radeon QD ");
			break;
		case PCI_DEVICE_ID_RADEON_QE:
			strcpy(rinfo->name, "Radeon QE ");
			break;
		case PCI_DEVICE_ID_RADEON_QF:
			strcpy(rinfo->name, "Radeon QF ");
			break;
		case PCI_DEVICE_ID_RADEON_QG:
			strcpy(rinfo->name, "Radeon QG ");
			break;
		case PCI_DEVICE_ID_RADEON_QY:
			strcpy(rinfo->name, "Radeon QY VE ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_QZ:
			strcpy(rinfo->name, "Radeon QZ VE ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_QW:
			strcpy(rinfo->name, "Radeon 7500 QW ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_QL:
			strcpy(rinfo->name, "Radeon 8500 QL ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_LW:
			strcpy(rinfo->name, "Radeon M7 LW ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_LY:
			strcpy(rinfo->name, "Radeon M6 LY ");
			rinfo->hasCRTC2 = 1;
			break;
		case PCI_DEVICE_ID_RADEON_LZ:
			strcpy(rinfo->name, "Radeon M6 LZ ");
			rinfo->hasCRTC2 = 1;
			break;
		default:
			return -ENODEV;
	}

	/* framebuffer size */
	tmp = INREG(CONFIG_MEMSIZE);

	/* mem size is bits [28:0], mask off the rest */
	rinfo->video_ram = tmp & CONFIG_MEMSIZE_MASK;

	/* ram type */
	tmp = INREG(MEM_SDRAM_MODE_REG);
	switch ((MEM_CFG_TYPE & tmp) >> 30) {
		case 0:
			/* SDR SGRAM (2:1) */
			strcpy(rinfo->ram_type, "SDR SGRAM");
			rinfo->ram.ml = 4;
			rinfo->ram.mb = 4;
			rinfo->ram.trcd = 1;
			rinfo->ram.trp = 2;
			rinfo->ram.twr = 1;
			rinfo->ram.cl = 2;
			rinfo->ram.loop_latency = 16;
			rinfo->ram.rloop = 16;
	
			break;
		case 1:
			/* DDR SGRAM */
			strcpy(rinfo->ram_type, "DDR SGRAM");
			rinfo->ram.ml = 4;
			rinfo->ram.mb = 4;
			rinfo->ram.trcd = 3;
			rinfo->ram.trp = 3;
			rinfo->ram.twr = 2;
			rinfo->ram.cl = 3;
			rinfo->ram.tr2w = 1;
			rinfo->ram.loop_latency = 16;
			rinfo->ram.rloop = 16;

			break;
		default:
			/* 64-bit SDR SGRAM */
			strcpy(rinfo->ram_type, "SDR SGRAM 64");
			rinfo->ram.ml = 4;
			rinfo->ram.mb = 8;
			rinfo->ram.trcd = 3;
			rinfo->ram.trp = 3;
			rinfo->ram.twr = 1;
			rinfo->ram.cl = 3;
			rinfo->ram.tr2w = 1;
			rinfo->ram.loop_latency = 17;
			rinfo->ram.rloop = 17;

			break;
	}

	rinfo->bios_seg = radeon_find_rom(rinfo);
	radeon_get_pllinfo(rinfo, rinfo->bios_seg);

	RTRACE("s3triofb: probed %s %dk videoram\n", (rinfo->ram_type), (rinfo->video_ram/1024));

#if !defined(__powerpc__)
	radeon_get_moninfo(rinfo);
#else
	switch (pdev->device) {
		case PCI_DEVICE_ID_RADEON_LW:
		case PCI_DEVICE_ID_RADEON_LY:
		case PCI_DEVICE_ID_RADEON_LZ:
			rinfo->dviDisp_type = MT_LCD;
			break;
		default:
			radeon_get_moninfo(rinfo);
			break;
	}
#endif

	radeon_get_EDID(rinfo);

	if ((rinfo->dviDisp_type == MT_DFP) || (rinfo->dviDisp_type == MT_LCD) ||
	    (rinfo->crtDisp_type == MT_DFP)) {
		if (!radeon_get_dfpinfo(rinfo)) {
			iounmap ((void*)rinfo->mmio_base);
			release_mem_region (rinfo->mmio_base_phys,
					    pci_resource_len(pdev, 2));
			release_mem_region (rinfo->fb_base_phys,
					    pci_resource_len(pdev, 0));
			kfree (rinfo);
			return -ENODEV;
		}
	}

	rinfo->fb_base = (u32) ioremap (rinfo->fb_base_phys,
				  		  rinfo->video_ram);
	if (!rinfo->fb_base) {
		printk ("s3triofb: cannot map FB\n");
		iounmap ((void*)rinfo->mmio_base);
		release_mem_region (rinfo->mmio_base_phys,
				    pci_resource_len(pdev, 2));
		release_mem_region (rinfo->fb_base_phys,
				    pci_resource_len(pdev, 0));
		kfree (rinfo);
		return -ENODEV;
	}

	/* XXX turn off accel for now, blts aren't working right */
	noaccel = 1;

	/* currcon not yet configured, will be set by first switch */
	rinfo->currcon = -1;

	/* set all the vital stuff */
	radeon_set_fbinfo (rinfo);

	/* save current mode regs before we switch into the new one
	 * so we can restore this upon __exit
	 */
	radeon_save_state (rinfo, &rinfo->init_state);

	/* init palette */
	for (i=0; i<16; i++) {
		j = color_table[i];
		rinfo->palette[i].red = default_red[j];
		rinfo->palette[i].green = default_grn[j];
		rinfo->palette[i].blue = default_blu[j];
	}

	pci_set_drvdata(pdev, rinfo);
	rinfo->next = board_list;
	board_list = rinfo;

	if (register_framebuffer ((struct fb_info *) rinfo) < 0) {
		printk ("s3triofb: could not register framebuffer\n");
		iounmap ((void*)rinfo->fb_base);
		iounmap ((void*)rinfo->mmio_base);
		release_mem_region (rinfo->mmio_base_phys,
				    pci_resource_len(pdev, 2));
		release_mem_region (rinfo->fb_base_phys,
				    pci_resource_len(pdev, 0));
		kfree (rinfo);
		return -ENODEV;
	}

	if (!noaccel) {
		/* initialize the engine */
		radeon_engine_init (rinfo);
	}

#ifdef CONFIG_PMAC_BACKLIGHT
	if (rinfo->dviDisp_type == MT_LCD)
		register_backlight_controller(&radeon_backlight_controller,
					      rinfo, "ati");
#endif

#ifdef CONFIG_PMAC_PBOOK
	if (rinfo->dviDisp_type == MT_LCD) {
		rinfo->pm_reg = pci_find_capability(pdev, PCI_CAP_ID_PM);
		pmu_register_sleep_notifier(&radeon_sleep_notifier);
	}
#endif

	printk ("s3triofb: ATI %s %s %d MB\n", rinfo->name, rinfo->ram_type,
		(rinfo->video_ram/(1024*1024)));

	if (rinfo->hasCRTC2) {
		printk("s3triofb: DVI port %s monitor connected\n",
			GET_MON_NAME(rinfo->dviDisp_type));
		printk("s3triofb: CRT port %s monitor connected\n",
			GET_MON_NAME(rinfo->crtDisp_type));
	} else {
		printk("s3triofb: CRT port %s monitor connected\n",
			GET_MON_NAME(rinfo->crtDisp_type));
	}

	RTRACE("s3triofb_pci_register END\n");
#endif
	return 0;
}



static void __devexit s3triofb_pci_unregister (struct pci_dev *pdev)
{
#if 0
        struct s3triofb_info *rinfo = pci_get_drvdata(pdev);
 
        if (!rinfo)
                return;
 
	/* restore original state */
	radeon_write_mode (rinfo, &rinfo->init_state);
 
        unregister_framebuffer ((struct fb_info *) rinfo);
                
        iounmap ((void*)rinfo->mmio_base);
        iounmap ((void*)rinfo->fb_base);
 
	release_mem_region (rinfo->mmio_base_phys,
			    pci_resource_len(pdev, 2));
	release_mem_region (rinfo->fb_base_phys,
			    pci_resource_len(pdev, 0));
        
        kfree (rinfo);
#endif
}


MODULE_LICENSE("GPL");


